# pylint: skip-file
from .basekeyset import BaseKeySet
from .keychain import Keychain
from .utils import AwaitableWrapper
from . import providers


__all__ = [
    'configure',
    'providers',
    'setup',
    'BaseKeySet',
    'Keychain'
]


_keychain: Keychain = Keychain()


def configure(spec: dict, keychain: Keychain = None) -> None:
    """Configure keys using a mapping of key names to their
    specifications.
    """
    (keychain or _keychain).register(spec)
    return AwaitableWrapper(setup, keychain=keychain)


def get_default_keychain() -> Keychain:
    """Return the default :class:`Keychain` instance."""
    return _keychain


async def ephemeral(**spec) -> providers.KeySpecification:
    """Load an ephemeral :class:`~ckms.ops.KeySpecification` with the
    given parameters, using the ``local`` provider.
    """
    spec = provider('local').load("noop", spec)
    return await spec


def provider(name: str, **kwargs) -> providers.Provider:
    """Get the provider for the backend `name`."""
    if kwargs: # pragma: no cover
        raise NotImplementedError
    return BaseKeySet.provider(name)


async def setup(keychain: Keychain = None) -> None:
    """Ensure that all key material and metadata is loaded for all keys
    that were previously registered in the :class:`Keychain`.
    """
    return await (keychain or _keychain).setup()


async def sign(data: bytes, using: str, algorithm: str) -> bytes:
    return await _keychain.sign(data, using, algorithm)


async def random(num_bytes: int):
    return await provider('random').random(num_bytes)


async def verify(**kwargs) -> bool:
    return await _keychain.verify(**kwargs)
