"""Declares :class:`EncryptionAlgorithm`."""
import typing

from .. import ops
from .algorithm import Algorithm


class EncryptionAlgorithm(Algorithm):
    __module__: str = 'ckms.algorithms'

    def __init__(self, wrap: bool = False):
        self.wrap = wrap


    def decrypt(
        self,
        provider: 'Provider',
        key: 'KeySpecification',
        data: typing.Union[bytes, ops.Ciphertext],
    ) -> bytes:
        """Perform a decryption operation. Return a byte-sequence holding the
        plain text.

        Construct a :class:`~ckms.ops.DecryptOperation` using the specified `key`,
        `data` and the parameters provider by the :class:`EncryptionAlgorithm`
        implementation.
        """
        if not isinstance(data, ops.Plaintext):
            data = ops.Ciphertext(data)
        op = self.get_decrypt_operation(
            data=data,
            **self.get_encryption_parameters(key, data)
        )
        return op.execute(provider, key)

    def encrypt(
        self,
        provider: 'Provider',
        key: 'KeySpecification',
        data: typing.Union[bytes, ops.Plaintext],
    ) -> bytes:
        """Perform an encryption operation. Return a byte-sequence holding the
        ciphertext.

        Construct a :class:`~ckms.ops.EncryptOperation` using the specified `key`,
        `data` and the parameters provider by the :class:`EncryptionAlgorithm`
        implementation.
        """
        if not isinstance(data, ops.Plaintext):
            data = ops.Plaintext(data)

        assert isinstance(data, ops.Plaintext)
        op = self.get_encrypt_operation(
            data=data,
            **self.get_encryption_parameters(key, data)
        )
        return op.execute(provider, key)

    def get_encryption_parameters(
        self,
        key: 'KeySpecification',
        data: bytes
    ) -> dict:
        """Hook to provide parameters to :class:`ckms.ops.EncryptOperation`."""
        return {}  # pragma: no cover

    def get_decrypt_operation(
        self,
        data: ops.Ciphertext,
        **kwargs
    ) -> typing.Union[ops.DecryptOperation]:
        """Instantiate the :class:`ckms.ops.Operation` implementation that
        is performed by this :class:`EncryptionAlgorithm`.
        """
        return ops.DecryptOperation(data, **kwargs)

    def get_encrypt_operation(
        self,
        data: ops.Plaintext,
        **kwargs
    ) -> typing.Union[ops.EncryptOperation, ops.KeyWrapOperation]:
        """Instantiate the :class:`ckms.ops.Operation` implementation that
        is performed by this :class:`EncryptionAlgorithm`.
        """
        cls = ops.EncryptOperation if not self.wrap else ops.KeyWrapOperation
        return cls(data, **kwargs)