"""Declares :class:`SigningAlgorithm`."""
import typing

from .. import ops
from .algorithm import Algorithm


class SigningAlgorithm(Algorithm):
    __module__: str = 'ckms.algorithms'

    def __init__(
        self,
        algorithm: typing.Literal['sha1', 'sha256', 'sha384', 'sha512'] = None,
    ):
        self._algorithm = algorithm

    def get_signing_parameters(
        self,
        key: 'KeySpecification',
        data: typing.Union[ops.Digest, ops.Message]
    ) -> dict:
        """Hook to provide parameters to :class:`ckms.ops.SigningOperation`."""
        return {}  # pragma: no cover

    def sign(
        self,
        provider: 'Provider',
        key: 'KeySpecification',
        data: typing.Union[ops.Digest, ops.Message]
    ) -> bytes:
        """Perform a signing operation. Return a byte-sequence holding the
        signature.

        Construct a :class:`~ckms.ops.SignOperation` using the specified `key`,
        `data` and the parameters provider by the :class:`SigningAlgorithm`
        implementation.
        """
        assert isinstance(data, (bytes, ops.Digest, ops.Message))
        if not isinstance(data, (ops.Digest, ops.Message)):
            data = ops.Message(data) # pragma: no cover
        op = ops.SignOperation(
            data=data,
            **self.get_signing_parameters(key, data)
        )
        return op.execute(provider, key)

    def verify(
        self,
        provider: 'Provider',
        key: 'KeySpecification',
        signature: bytes,
        data: typing.Union[ops.Digest, ops.Message]
    ) -> bytes:
        """Perform a signing operation. Return a byte-sequence holding the
        signature.

        Construct a :class:`~ckms.ops.SignOperation` using the specified `key`,
        `data` and the parameters provider by the :class:`SigningAlgorithm`
        implementation.
        """
        if not isinstance(data, (ops.Digest, ops.Message)):
            data = ops.Message(data) # pragma: no cover
        op = ops.VerifyOperation(
            signature=signature,
            data=data,
            **self.get_signing_parameters(key, data)
        )
        return op.execute(provider, key)