"""Declares :class:`BaseKeySet`."""
import asyncio
import collections
import itertools
import typing

from . import algorithms
from . import ops
from . import providers
from .keymetadata import KeyMetadata


class BaseKeySet:
    """Specifies the basic interface of a keyset implementaton, providing
    methods to perform cryptographic operations using named keys.
    """
    __module__: str = 'ckms'
    _providers: collections.OrderedDict[str, providers.Provider] = collections.OrderedDict()

    @staticmethod
    def provider(name: str) -> providers.Provider:
        """Return a :class:`ckms.providers.Provider` instance that was
        priorly registered under `name`.
        """
        if name not in BaseKeySet._providers: # pragma: no cover
            BaseKeySet._providers[name] = providers.get(name)
        return BaseKeySet._providers[name]

    def __init__(self):
        self._aliases = {}
        self._keys = collections.OrderedDict()
        self._trust = collections.OrderedDict()
        self._providers = BaseKeySet._providers
        self._tags = collections.defaultdict(list)
        self._public = {}

        # Ensure that a random provider is configured
        if 'random' not in self._providers:
            self._providers['random'] = providers.RandomProvider()

    def add(self, spec: providers.KeySpecification) -> None:
        """Add a key described by :class:`KeySpecification` `spec` to the
        internal key registry.
        """
        self._keys[spec.kid] = spec
        if hasattr(spec, 'tags'): # TODO
            for tag in spec.tags:
                self._tags[tag].append(spec)

    def alias(self, name: str, alias: str) -> None:
        """Create an alias for key `name`."""
        self._aliases[alias] = name

    def get(self, kid: str, public: bool = False) -> providers.KeySpecification:
        """Return the key configuration identified by `kid`."""
        return (self._trust if public else {}).get(kid) or self._keys[kid]

    def get_algorithm(self, name: str) -> algorithms.Algorithm:
        """Return the :class:`ckms.algorithms.Algorithm` instance specified by
        the `name` parameter.
        """
        return algorithms.get(name)

    def get_decryption_key(self, kid: str) -> providers.KeySpecification:
        """Lookup a decryption key using a key identifier."""
        if kid in self._trust:
            public = self.get(kid, public=True)
            kid = public.get_private_kid()
        if kid in self._aliases:
            kid = self._aliases[kid]
        return self.get(kid)

    def get_verification_key(self, kid: str) -> providers.KeySpecification:
        """Lookup a verification key using a key identifier."""
        return self._public.get(kid) or self._keys[kid]

    def get_key_material(self, name: str) -> typing.Tuple[bytes, bytes]:
        """Return the cryptographic key material, if available. This method
        is mostly used during test cases; production software should not
        access raw key material.
        """
        return self.get(name).get_key_material()

    def get_key_metadata(self, name: str) -> KeyMetadata:
        """Return a metadata instance describing a the key `name`."""
        return KeyMetadata(**self.get(name).dict())

    def get_provider(self, name: str) -> providers.Provider:
        """Return a :class:`ckms.providers.Provider` instance that was
        priorly registered under `name`.
        """
        if name not in self._providers:
            self._providers[name] = providers.get(name)
        return self._providers[name]

    def get_public_key(self, kid: str) -> 'KeySpecification':
        """Return a priorly registered public key."""
        return self._trust[kid]

    def get_public_kid(self, name: str) -> str:
        """Return the public key identifier of the given key `name`."""
        return self._keys[name].public_kid

    def get_public_key_versions(
        self,
        kid: str,
        factory: typing.Callable[[object], object] = None
    ) -> typing.List[typing.Any]:
        """Return the list of public keys for the given kid."""
        if factory is None:
            factory = lambda x: x
        key = self.get(kid)
        return [factory(v) for v in key.get_public_keys()]

    def get_by_tag(self, tag: str) -> typing.List[providers.KeySpecification]:
        """Return all key identifiers for the given `tag`."""
        return {x for x in self._tags[tag]}

    def get_tagged(self, tags: typing.List[str]) -> typing.List[str]:
        """Return a key identifiers for the keys that are tagged with
        any of`tags`.
        """
        keys = set(itertools.chain(*map(self.get_by_tag, tags)))
        return [x.kid for x in sorted(keys, key=lambda x: x.weight)]

    def has_public_key(self, kid: str) -> bool:
        """Return a boolean indicating if the public key identified
        by `kid` is known.
        """
        return kid in self._keys or kid in self._aliases

    def load(
        self,
        provider_name: str,
        kid: str,
        spec: dict
    ) -> providers.KeySpecification:
        """Instantiate a :class:`ckms.providers.KeySpecification` instance
        using the given parameters.
        """
        provider = self.get_provider(provider_name)
        return provider.load(kid, spec)

    def register(self, keys: dict[str, typing.Any]) -> None:
        """Registers a mapping of key names to specifications."""
        for kid, spec in dict.items(keys): # pragma: no cover
            if not spec.get('provider'):
                raise ValueError("The specification must declare the `provider` member.")
            if spec['provider'] not in providers.get_available(): # pragma: no cover
                raise ValueError(f"Unknown provider: {spec['provider']}")
            spec.setdefault('keychain', self)
            self.add(self.load(spec.pop('provider'), kid, spec))

    def register_public_key(self, public: providers.KeySpecification):
        """Registers public key `public` for the key `name`."""
        self._public[public.kid] = public

    def trust(self, spec: 'KeySpecification') -> None:
        """Trust a public key using the ``local`` provider."""
        self._trust[spec.kid] = spec

    def decrypt(
        self,
        data: typing.Union[ops.Ciphertext, str],
        using: str,
        algorithm: str
    ) -> bytes:
        """Decrypt `data` using the given preregistered key `using` and
        `algorithm`.
        """
        alg = self.get_algorithm(algorithm)
        key = self.get_decryption_key(using)
        return alg.decrypt(key.provider, key, data)

    def encrypt(
        self,
        data: typing.Union[ops.Plaintext, str],
        using: str,
        algorithm: str
    ) -> ops.Ciphertext:
        """Encrypt `data` using the given preregistered key `using` and
        `algorithm`.
        """
        alg = self.get_algorithm(algorithm)
        key = self.get(using)
        return alg.encrypt(key.provider, key, data)

    async def setup(self):
        """Ensure that all key material and metadata is loaded for all keys
        that were previously registered in the :class:`BaseKeySet`.
        """
        futures = []
        for spec in dict.values(self._keys):
            if spec.is_loaded(): # pragma: no cover
                continue
            futures.append(spec.load())
        await asyncio.gather(*futures)

    def sign(
        self,
        data: typing.Union[bytes, ops.Digest, ops.Message],
        using: str,
        algorithm: str
    ) -> typing.Union[bytes, typing.Awaitable[bytes]]:
        """Create a digital signature using a preregistered key.

        The `data` argument is the input data for the digital signature. If no
        encoding and algorithm is forced, then these are automatically discovered.

        The `using` argument specifies the key to use. It must be preregistered
        in the keychain.

        With the `algorithm` argument, the concrete signing algorithm is
        chosen. Legal values for `algorithm` are defined in :rfc:`7518`
        **JSON Web Algorithms (JWA)** and the IANA JSON Object Signing and
        Encryption (JOSE) registry.
        """
        if isinstance(data, bytes):
            data = ops.Message(data)
        alg = self.get_algorithm(algorithm)
        key = self.get(using)
        return alg.sign(key.provider, key, data)

    def verify(
        self,
        *,
        signature: bytes,
        data: bytes,
        using: str,
        algorithm: str
    ) -> bool:
        """Return a boolean indicating if `signature` is valid for `data."""
        if isinstance(data, bytes):
            data = ops.Message(data)
        alg = self.get_algorithm(algorithm)
        key = self.get_verification_key(using)
        return alg.verify(key.provider, key, signature, data)

    def __iter__(self):
        return iter(self._keys.values())
