"""Declares :class:`LocalProvider`."""
from typing import Literal

import aiofiles

from ckms import core
from ckms.core.models import KeySpecification
from ckms.core.provider import Provider
from ckms.core.types import KeyOperationType
from .models import LocalKey
from .models import TransientKey


class LocalProvider(Provider):
    """A :class:`~ckms.core.Provider` implementation that used the
    local CPU to perform cryptographic operations.
    """
    require_algorithm: bool = True
    supported_ed_curves: set[str] = {"Ed448", "X448", "Ed25519", "X25519"}

    class spec(KeySpecification):
        provider: Literal['local'] = 'local'
        use: KeyOperationType
        key: LocalKey | TransientKey | None

    async def fetch(self, params: core.RemoteBlob) -> bytes:
        """Fetches the content of the remote blob using the parameters
        specified. Return a byte-string holding the content.
        """
        blob = LocalKey.parse_obj(params.dict())
        async with aiofiles.open(blob.path, 'rb') as f: # type: ignore
            return await f.read() # type: ignore

    def validate_spec(self, spec: spec):
        if spec.can_sign() and spec.kty == "OKP":
            # OKP signing keys always use the EdDSA algorithm.
            spec.algorithm = "EdDSA"
        if self.require_algorithm and not spec.algorithm:
            raise ValueError(
                "The `algorithm` parameter is required for provider "
                f"{type(self).__name__}"
            )
        if spec.kty in {"EC", "OKP"}\
        and spec.curve not in self.supported_ed_curves:
            if spec.curve is None:
                raise ValueError(
                    "The `curve` parameter is required when using keys of type "
                    "EC or OKP."
                )
            else:
                raise ValueError(f"Invalid curve: {spec.curve}")
        return spec