# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`KeySpecification`."""
import pydantic

from ckms.core.types import KeyOperationType


class KeySpecification(pydantic.BaseModel):
    """The base class for all provider key configurations."""
    __module__: str = 'ckms.core.models'
    provider: str
    kty: str
    kid: str | None = None
    use: KeyOperationType | None = None
    algorithm: str | None = None
    curve: str | None = None
    loaded: bool = False
    tags: set[str] = set()
    public: bytes | None = None

    def can_encrypt(self) -> bool:
        """Return a boolean indicating if the key may be used to
        encrypt.
        """
        return self.use == "enc"

    def can_sign(self) -> bool:
        """Return a boolean indicating if the key may be used to
        create digital signatures.
        """
        return self.use == "sig"

    def is_asymmetric(self) -> bool:
        """Return a boolean indicating if the key is an asymmetric
        algorithm.
        """
        return self.kty in {"EC", "OKP", "RSA"}

    def is_symmetric(self) -> bool:
        """Return a boolean indicating if the key is a symmetric
        algorithm.
        """
        return self.kty in {"oct"}

    def is_loaded(self) -> bool: # pragma: no cover
        """Return a boolean indicating if the metadata describing the
        key is retrieved.
        """
        return self.loaded