"""Declares :class:`Provider`."""
import importlib
from typing import cast
from typing import Any
from typing import Awaitable
from types import UnionType

from .models import KeySpecification
from .keyinspector import KeyInspector
from .types import IProvider
from .utils import b64encode


class Provider(IProvider):
    __module__: str = 'ckms.core'
    _default_providers: dict[str, str] = {
        'google': 'ckms.ext.google.GoogleProvider',
        'local': 'ckms.core.local.LocalProvider',
        'random': 'ckms.core.random.RandomProvider',
        'ssh-agent': 'ckms.ext.ssh.provider.SSHAgentProvider',
    }
    _providers: dict[str, IProvider] = {}
    inspector: KeyInspector
    name: str
    spec: type[KeySpecification] | UnionType | None = None

    @staticmethod
    def get(name: str) -> IProvider:
        """Return the provider implementation specified by the `name`
        argument.
        """
        if name not in Provider._providers:
            ProviderImpl = cast(type[Provider], Provider.load_impl(name))
            Provider._providers[name] = ProviderImpl()
            if ProviderImpl.spec is not None:
                Provider.spec |= ProviderImpl.spec
        return Provider._providers[name]

    @staticmethod
    def load_impl(name: str) -> type[IProvider]:
        """Loads the implementation class `name`."""
        module_name, symbol = str.rsplit(Provider._default_providers[name], '.', 1)
        return getattr(importlib.import_module(module_name), symbol)

    @staticmethod
    def register(name: str, qualname: str) -> None:
        """Registers a new provider implementation under the given name."""
        Provider._default_providers[name] = qualname # pragma: no cover

    def __init__(
        self,
        inspector: KeyInspector | None = None
    ):
        self._providers = Provider._providers
        self.inspector = inspector or KeyInspector()
        self.on_init()

    def on_init(self) -> None:
        """Called when the :class:`Provider` is initialized."""
        pass

    def configure(self, **params: Any) -> None:
        """Configures the provider with the given parameters."""
        raise NotImplementedError

    async def fetch(self, params: Any) -> bytes:
        """Fetches the content of the remote blob using the parameters
        specified. Return a byte-string holding the content.
        """
        raise NotImplementedError

    async def random(self, length: int) -> bytes:
        """Return a random byte-sequence of the given length."""
        random = self.get('random')
        return await random.random(length=length)

    async def random_urlsafe(self, length: int) -> str:
        """Like :meth:`random()`, but returs a Base64, URL-safe encoded
        string.
        """
        return bytes.decode(b64encode(await self.random(length)))

    def sign(
        self,
        spec: KeySpecification
    ) -> bytes | Awaitable[bytes]:
        """Sign the given digest or message with the specified key. Return
        a byte-sequence holding the signature.
        """
        raise NotImplementedError