# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from typing import cast

import google.auth
from google.auth.credentials import Credentials
from google.cloud.secretmanager import SecretManagerServiceAsyncClient

from ckms import core
from .models import SecretManagerBlob


class GoogleProvider(core.Provider):
    __module__: str = 'ckms.ext.google'
    credentials: Credentials
    project: str

    def on_init(self):
        self.credentials, self.project = cast(Credentials, google.auth.default())
        self.secret_manager = SecretManagerServiceAsyncClient(
            credentials=self.credentials
        )

    async def fetch(self, params: core.RemoteBlob) -> bytes:
        """Fetches the content of the remote blob using the parameters
        specified. Return a byte-string holding the content.
        """
        blob = SecretManagerBlob.parse_obj(params.dict())
        client = SecretManagerServiceAsyncClient(credentials=self.credentials)
        secret = client.secret_path(blob.project, blob.secret)
        async with client:
            response = await client.access_secret_version(
                request={
                    'name': f'{secret}/versions/{blob.version}'
                }
            )
        return response.payload.data