"""Declares :class:`FixedAlgorithmKey`"""
import typing

import ckms
from .algorithms import Algorithm
from .ops import Ciphertext
from .ops import Plaintext
from . import providers


KEY_LENGTHS = {
    "A128CBC-HS256" : 16,
    "A192CBC-HS384" : 24,
    "A256CBC-HS512" : 32,
    "A128GCM"       : 16,
    "A192GCM"       : 24,
    "A256GCM"       : 32,
}


class FixedAlgorithmKey:
    """An interface to public, private or symmetric keys, that
    performs cryptographic operations conforming to a preset
    algorithm.
    """
    __module__: str = 'ckms'

    @classmethod
    async def generate(
        cls,
        algorithm: str,
        random: typing.Optional[providers.Provider]
    ) -> 'FixedAlgorithmkey':
        """Generate a new content encryption key (CEK) using the
        specified entropy provider.
        """
        spec = await ckms.ephemeral(
            secret=await random.random(KEY_LENGTHS[algorithm]),
            use="enc"
        )
        return cls(algorithm, spec)

    def __init__(
        self,
        algorithm: typing.Union[str, Algorithm],
        spec: providers.KeySpecification
    ):
        if isinstance(algorithm, str):
            algorithm = Algorithm.get(algorithm)
        self.algorithm = algorithm
        self.spec = spec

    async def encrypt(
        self,
        pt: typing.Union[bytes, Plaintext],
        aad: typing.Optional[bytes] = None
    ) -> Ciphertext:
        if not isinstance(pt, Plaintext):
            pt = Plaintext(pt, aad=aad)
        return await self.algorithm.encrypt(
            provider=self.spec.provider,
            key=self.spec,
            data=pt
        )

    async def decrypt(self, ct: Ciphertext) -> bytes:
        return await self.algorithm.decrypt(
            provider=self.spec.provider,
            key=self.spec,
            data=ct
        )

    def __bytes__(self) -> bytes:
        return bytes(self.spec.key)
