"""Declares :class:`ClaimSet`."""
import datetime
import json
import secrets
import typing
from typing import cast
from typing import Any
from typing import Callable

import pydantic

from ckms.utils import b64encode
from ckms.utils import b64decode
from ckms.utils import current_timestamp
from .audiencetype import AudienceType
from .exceptions import Expired
from .exceptions import MaximumAgeExceeded
from .exceptions import InvalidClaimType
from .exceptions import MalformedPayload
from .exceptions import NotEffective
from .exceptions import MissingRequiredClaim
from .exceptions import WrongAudience


class ClaimSet(pydantic.BaseModel):
    __module__: str = 'ckms.jose'
    jti: typing.Optional[typing.Any]
    iss: typing.Optional[str]
    aud: AudienceType = AudienceType()
    sub: typing.Optional[typing.Union[int, str]]
    iat: typing.Optional[int]
    nbf: typing.Optional[int]
    exp: typing.Optional[int]
    extra: typing.Dict[str, typing.Any] = {}

    @classmethod
    def frompayload(cls, payload: bytes) -> 'ClaimSet':
        try:
            claims = json.loads(b64decode(payload))
        except (TypeError, ValueError):
            raise MalformedPayload
        return cls(**claims)

    @classmethod
    def fromb64(cls, payload: bytes) -> 'ClaimSet':
        return cls.frompayload(payload)

    @classmethod
    def strict(
        cls,
        *,
        iss: str,
        aud: typing.Union[str, typing.Set[str]],
        ttl: int,
        random: typing.Callable[[int], typing.Union[int, str]] = secrets.token_urlsafe,
        **claims: str
    ) -> 'ClaimSet':
        """Return a :class:`ClaimSet` and ensure that the ``iss``, ``aud``,
        ``iat``, ``nbf``, ``exp`` and ``jti`` claims are set.
        """
        now = current_timestamp()
        return cls(
            jti=random(24),
            iss=iss,
            aud=aud,
            iat=now,
            nbf=claims.pop('nbf', None) or now,
            exp=now+ttl,
            **claims
        )

    def __init__(self, **claims: typing.Any):
        extra = claims.setdefault('extra', {})
        extra.update({
            x: claims.pop(x)
            for x in list(claims.keys())
            if x not in self.__fields__
        })
        try:
            super().__init__(**claims)
        except pydantic.ValidationError as exception:
            for error in exception.errors():
                error_type = error.get('type') or ''
                if str.startswith(error_type, 'type_error'):
                    claim = error['loc'][0]
                    raise InvalidClaimType(str(claim)) from exception
            else:
                raise

    def expires_in(
        self,
        ttl: typing.Optional[int] = None,
        now: typing.Optional[int] = None
    ) -> typing.Optional[int]:
        """Set the time-to-live using the ``exp`` claim. Return
        the remaining seconds.
        """
        now = now if now is not None else current_timestamp()
        if ttl is not None:
            self.exp = now + ttl
        return (self.exp - now) if self.exp else None

    def is_selfsigned(self) -> bool:
        """Return a boolean indicating if the claims are self-signed."""
        return self.iss == self.sub

    def verify(
        self,
        audience: typing.Optional[typing.Set[str]] = None,
        required: typing.Optional[typing.Set[str]] = None,
        now: typing.Optional[int] = None,
        max_age: typing.Optional[int] = None,
        strict: bool = False
    ) -> None:
        """Verifies the claims according to the specification laid out
        in the JOSE RFCs.
        """
        now = now or current_timestamp()
        audience = audience or set()
        required = required or set()
        if strict:
            required.update({'iat', 'nbf', 'exp'})
        if max_age:
            required.add('iat')
        for claim in required:
            if getattr(self, claim, self.extra.get(claim)) is not None:
                continue
            raise MissingRequiredClaim(claim)
        if self.exp is not None and self.exp <= now:
            raise Expired
        if self.nbf and self.nbf > now:
            raise NotEffective
        if audience and not bool(audience & self.aud):
            raise WrongAudience(allowed=audience)
        if max_age is not None and (now - cast(int, self.iat)) >= max_age:
            raise MaximumAgeExceeded

    def dict( # type: ignore
        self,
        *args: typing.Any,
        **kwargs: typing.Any
    ) -> typing.Dict[str, typing.Any]:
        kwargs.setdefault('exclude_defaults', True)
        data = super().dict(*args, **kwargs)
        data.update(data.pop('extra', {}))
        return data

    def json( # type: ignore
        self,
        *args: typing.Any,
        **kwargs: typing.Any
    ) -> str:
        kwargs.setdefault('exclude_defaults', True)
        encoder = cast(
            Callable[[Any], Any],
            kwargs.pop('encoder', None) or self.__json_encoder__
        )
        return self.__config__.json_dumps(
            obj=self.dict(**kwargs),
            default=encoder
        )

    def __bytes__(self) -> bytes:
        return b64encode(str.encode(self.json(exclude_defaults=True), 'utf-8'))

    def __getattr__(self, attname: str) -> typing.Any:
        return self.extra.get(attname) if self.extra else None