"""Declares exception classes related to JOSE errors."""
import typing

from unimatrix.exceptions import CanonicalException


class JOSEException(CanonicalException):
    http_status_code: int = 403


class Malformed(JOSEException):
    """Raised when an aspect of a JOSE object is malformed."""
    __module__: str = 'ckms.jose'
    code: str = 'MALFORMED_OBJECT'


class MalformedToken(Malformed):
    """Raised when a token is malformed."""
    __module__: str = 'ckms.jose'
    code: str = 'MALFORMED_TOKEN'


class MalformedHeader(Malformed):
    """Raised when the header of a JOSE object could not be
    decoded.
    """
    __module__: str = 'ckms.jose'
    code: str = 'MALFORMED_HEADER'
    message: str = (
        "The protected JOSE header was malformed."
    )


class MalformedPayload(Malformed):
    """Raised when the payload that is expected to be a JWT could
    not be decoded.
    """
    __module__: str = 'ckms.jose'
    code: str = 'MALFORMED_PAYLOAD'
    message: str = "The JOSE payload was malformed."


class InvalidHeader(MalformedToken):
    """Raised when the header of a JOSE object is semantically
    invalid.
    """
    __module__: str = 'ckms.jose.exceptions'
    code: str = 'INVALID_JOSE_HEADER'


class InvalidSignature(JOSEException):
    """Raised when one or more of the signatures on a JSON Web
    Signature (JWS) did not validate.
    """
    __module__: str = 'ckms.jose.exceptions'
    code: str = 'INVALID_SIGNATURE'


class AlgorithmNotAllowed(InvalidSignature):
    __module__: str = 'ckms.jose.exceptions'
    code: str = 'INVALID_SIGNATURE_ALGORITHM'


class MisconfiguredIssuer(JOSEException):
    """Raised when the issuer metadata could be retrieved, but it
    was somehow misconfigured.
    """
    __module__: str = 'ckms.jose.exceptions'
    http_status_code: int = 503
    code: str = 'OAUTH2_ISSUER_NOT_AVAILABLE'


class UnsupportedAlgorithm(InvalidHeader):
    """Raised when the ``alg`` claim specifies an unsupported
    algorithm.
    """
    __module__: str = 'ckms.jose.exceptions'
    code: str = 'UNSUPPORTED_ALGORITHM'

    def __init__(self, given: str, supported: list):
        self.given = given
        self.supported = supported
        self.message = (
            f"Unsupported algorithm specified by the `alg` claim: {given}. "
            f"Supported algorithms are: {list(sorted(supported))}"
        )
        super().__init__(message=self.message)


class UnsupportedEncryption(InvalidHeader):
    """Raised when the ``enc`` claim specifies an unsupported
    algorithm.
    """
    __module__: str = 'ckms.jose.exceptions'
    code: str = 'UNSUPPORTED_CONTENT_ENCRYPTION'

    def __init__(self, given: str, supported: list):
        self.given = given
        self.supported = supported
        self.message = (
            f"Unsupported encryption specified by the `enc` claim: {given}. "
            f"Supported encryption algorithms are: {list(sorted(supported))}"
        )
        super().__init__(message=self.message)


class InvalidClaim(JOSEException):
    """Raised when the claim on a JSON Web Token (JWT)
    is invalid or illegal.
    """
    __module__: str = 'ckms.jose'
    code: str = 'ILLEGAL_CLAIM'

    def __init__(self, claim: str, detail: typing.Optional[str] = None):
        super().__init__(
            message=(
                f"The `{claim}` claim declared on the JSON Web Token "
                "(JWT) was illegal, invalid or of an improper type."
            ),
            detail=detail
        )


class MissingRequiredClaim(InvalidClaim):
    """Raises when the specified claim is not present but was
    required.
    """
    __module__: str = 'ckms.jose'
    code: str = 'MISSING_REQUIRED_CLAIM'


class InvalidClaimType(InvalidClaim):
    """Raised when the claim on a JSON Web Token (JWT)
    is of the wrong type.
    """
    __module__: str = 'ckms.jose'
    code: str = 'ILLEGAL_CLAIM_TYPE'


class Unusable(JOSEException):
    """Raised when a token is not usable."""
    __module__: str = 'ckms.jose.exceptions'


class Undecryptable(Unusable):
    """Raised when a received JWE is not decryptable."""
    __module__: str = 'ckms.jose.exceptions'
    code: str = 'UNDECRYPTABLE'


class Unverifiable(Unusable):
    """Raised when a received JWS is not verifiable."""
    __module__: str = 'ckms.jose.exceptions'
    code: str = 'UNVERIFIABLE'

    def __init__(self, kid: str):
        super().__init__(
            message=(
                "The signature could not be verified because the key "
                "specified by the `kid` claim is not known."
            )
        )


class InvalidToken(JOSEException):
    """Raised when a JSON Web Token (JWT) is not valid based
    on its claims.
    """
    __module__: str = 'ckms.jose'


class Expired(InvalidToken):
    """Raised when a JWT can not be used, based on the ``exp`` claim."""
    __module__: str = 'ckms.jose'
    code: str = 'TOKEN_EXPIRED'


class MaximumAgeExceeded(InvalidToken):
    """Raised when the maximum age of a JWT is exceeded based on the
    ``max_age`` parameter at validation time.
    """
    __module__: str = 'ckms.jose'
    code: str = 'TOKEN_MAX_AGE_EXCEEDED'


class NotEffective(InvalidToken):
    """Raised when a JWT can not be used, based on the ``nbf`` claim."""
    __module__: str = 'ckms.jose'
    code: str = 'TOKEN_NOT_EFFECTIVE'
    http_status_code: int = 401


class UnsupportedContentType(InvalidToken):
    """Raised when a JWE contains an unsupported content type."""
    __module__: str = 'ckms.jose'
    code: str = 'UNSUPPORTED_MEDIA_TYPE'
    http_status_code: int = 415


class UnsupportedTokenType(InvalidToken):
    """Raised when a JWT is of a type that is not accepted."""
    __module__: str = 'ckms.jose'
    code: str = 'UNSUPPORTED_TOKEN_TYPE'
    http_status_code: int = 415


class WrongAudience(InvalidToken):
    """Raised when none of the audiences in the ``aud`` claim intersected
    with the set of allowed audiences.
    """
    __module__: str = 'ckms.jose'
    code: str = "WRONG_AUDIENCE"

    def __init__(self, allowed: typing.Set[str]):
        super().__init__(
            message=(
                "The `aud` did not specify any acceptable audience. "
                "Present a token for any of the following audiences: "
                f"{str.join(', ', sorted(allowed))}"
            )
        )
