"""Declares :class:`JOSEDeserializer`."""
import typing

from ckms.utils import b64decode
from .exceptions import MalformedToken
from .jsonwebencryption import JSONWebEncryption
from .jsonwebsignature import JSONWebSignature
from .jsonwebtoken import JSONWebToken
from .signature import Signature


class JOSEDeserializer:
    __module__: str = 'ckms.jose'

    def deserialize_compact(
        self,
        token: typing.Union[bytes, str],
        using: str = None
    ) -> typing.Union[JSONWebSignature, JSONWebEncryption]:
        """Deserialize a JSON Web Token (JWT) from either a JSON Web
        Signature (JWS) or JSON Web Encryption (JWE).
        """
        if not isinstance(token, (bytes, str)): # pragma: no cover
            raise TypeError(f"Invalid `token` parameter: {repr(token)}")
        if isinstance(token, str): # pragma: no cover
            token = str.encode(token, "utf-8")
        protected, *parts = bytes.split(token, b'.')

        # Try to guess the type based on the number of parts. If the
        # number of parts is not 3 or 5, then the token is malformed.
        # If the number of parts is 3, then the token is a JSON Web
        # Signature (JWS), else JSON Web Encryption()
        if len(parts) not in {2, 4}:
            raise MalformedToken
        return self.deserialize_compact_jws(protected, *parts)\
            if len(parts) == 2\
            else self.deserialize_compact_jwe(protected, *parts)

    def deserialize_compact_jwe(
        self,
        protected: bytes,
        key: bytes,
        iv: bytes,
        ct: bytes,
        tag: bytes
    ) -> JSONWebEncryption:
        """Deserialize a JSON Web Encryption (JWE) object."""
        return JSONWebEncryption(
            protected=protected,
            key=key,
            iv=iv,
            ct=ct,
            tag=tag
        )

    def deserialize_compact_jws(
        self,
        protected: bytes,
        payload: bytes,
        signature: bytes
    ) -> JSONWebSignature:
        """Deserialize a JSON Web Signature (JWS)."""
        return JSONWebSignature(
            payload=payload,
            signatures=[Signature(
                protected=protected,
                signature=b64decode(signature)
            )]
        )

    def __repr__(self) -> str: # pragma: no cover
        return "JOSEDeserializer()"
