"""Declares :class:`JOSEHeader`."""
import binascii
import json
import typing

import pydantic

from ckms import ops
from ckms.utils import b64decode
from ckms.utils import b64encode
from .exceptions import MalformedHeader
from .exceptions import InvalidHeader
from .exceptions import UnsupportedAlgorithm
from .exceptions import UnsupportedEncryption


class JOSEHeader(pydantic.BaseModel):
    """Provides an interface do decode and encode the header of
    a JSON Web Signature (JWS) or JSON Web Encryption (JWE).
    """
    __module__: str = 'ckms.jose'
    _allowed_signing_algorithms: typing.Set[str] = {
        "EdDSA",
        "HS256",
        "HS384",
        "HS512",
        "RS256",
        "RS384",
        "RS512",
        "ES256",
        "ES384",
        "ES512",
        "ES256K",
        "PS256",
        "PS384",
        "PS512",
        "none"
    }

    _allowed_encryption_algorithms: typing.Set[str] = {
        "RSA1_5",
        "RSA-OAEP",
        "RSA-OAEP-256",
        "RSA-OAEP-384",
        "RSA-OAEP-512",
        "A128KW",
        "A192KW",
        "A256KW",
        "dir",
        "ECDH-ES",
        "ECDH-ES+A128KW",
        "ECDH-ES+A192KW",
        "ECDH-ES+A256KW",
        "A128GCMKW",
        "A192GCMKW",
        "A256GCMKW",
        "PBES2-HS256+A128KW",
        "PBES2-HS384+A192KW",
        "PBES2-HS512+A256KW",
    }

    _allowed_content_encryption_algorithms: typing.Set[str] = {
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512",
        "A128GCM",
        "A192GCM",
        "A256GCM",
    }


    alg: typing.Literal[tuple(_allowed_signing_algorithms | _allowed_encryption_algorithms)]
    jku: typing.Optional[str] = None
    jwk: typing.Optional[typing.Dict[str, typing.Any]] = {}
    kid: typing.Optional[str] = None
    x5u: typing.Optional[str] = None
    x5c: typing.Optional[str] = None
    x5t: typing.Optional[str] = None
    x5t_s256: typing.Optional[str] = pydantic.Field(None, alias='x5t#S256')
    typ: typing.Optional[str] = None
    cty: typing.Optional[str] = None
    crit: typing.Optional[typing.Set[str]] = set()
    enc: typing.Optional[typing.Literal[tuple(_allowed_content_encryption_algorithms)]] = None
    zip: typing.Optional[str] = None

    # AES GCM JWE
    iv: typing.Optional[str] = None
    tag: typing.Optional[str] = None

    @classmethod
    def fromb64(cls, encoded: typing.Union[bytes, str]) -> 'JOSEHeader':
        """Decode a :class:`JOSEHeader` instance from am
        URL-encoded, base64 byte-sequence.
        """
        if isinstance(encoded, str): # pragma: no cover
            encoded = str.encode(encoded, "utf-8")
        try:
            header = json.loads(b64decode(encoded))
        except (binascii.Error, TypeError, ValueError):
            raise MalformedHeader
        return cls.fromdict(header)

    @classmethod
    def fromcek(cls, cek: ops.Ciphertext, claims: dict) -> 'JOSEHeader':
        """Create a new :class:`JOSEHeader` from an encrypted content
        encryption key (CEK).
        """
        return cls.fromdict(claims)

    @classmethod
    def fromdict(cls, header: dict) -> 'JOSEHeader':
        """Create a new :class:`JOSEHeader` from a dictionary."""
        try:
            header = cls(**header)
        except pydantic.ValidationError as exc:
            for error in exc.errors():
                typ = error.get('type')
                if typ == "value_error.const" and 'alg' in error['loc']:
                    raise UnsupportedAlgorithm(
                        given=error['ctx']['given'],
                        supported=error['ctx']['permitted']
                    )
                if typ == "value_error.const" and 'enc' in error['loc']:
                    raise UnsupportedEncryption(
                        given=error['ctx']['given'],
                        supported=error['ctx']['permitted']
                    )

            raise InvalidHeader
        return header

    def is_encrypted(self) -> bool:
        """Return a boolean indicating if the header is encrypted."""
        return self.enc is not None

    def __bytes__(self) -> bytes:
        return b64encode(
            str.encode(self.json(exclude_defaults=True, by_alias=True), "utf-8")
        )
