"""Declares :class:`JSONWebEncryption`."""
import typing

from ckms.ops import Ciphertext
from ckms.utils import b64decode
from .joseheader import JOSEHeader
from .josepayloadcontainer import JOSEPayloadContainer


class JSONWebEncryption(JOSEPayloadContainer):
    __module__: str = 'ckms.jose'

    @property
    def header(self) -> JOSEHeader:
        """The protected JOSE header."""
        return JOSEHeader.fromb64(self._protected)

    def __init__(
        self,
        protected: bytes,
        key: bytes,
        iv: bytes,
        ct: bytes,
        tag: bytes
    ):
        self._protected = protected
        self._key = key
        self._iv = iv
        self._ct = ct
        self._tag = tag

    async def get_payload(
        self,
        codec: 'PayloadCodec',
        verify: bool = True,
        allow_none: bool = False,
    ) -> typing.Union[bytes, str, JOSEPayloadContainer]:
        """Decrypt the payload of the :class:`JSONWebEncryption`."""
        cek = await codec.decrypt_cek(
            header=self.header,
            ct=Ciphertext(
                data=b64decode(self._key),
                aad=b'',
                tag=b64decode(self._tag or b'') or None,
                iv=b64decode(self._iv or b'') or None
            )
        )
        payload = await cek.decrypt(
            Ciphertext(
                data=b64decode(self._ct),
                aad=self._protected,
                iv=b64decode(self._iv),
                tag=b64decode(self._tag)
            )
        )
        if self.header.cty == "JWT":
            # Presumably a JWS.
            payload = await codec.decode(payload)
        elif self.header.typ == "JWT":
            payload = codec.deserialize_jwt(payload)
        else:
            payload = b64decode(payload)
        return payload

    def __bytes__(self) -> bytes:
        return bytes.join(b'.', [
            self._protected,
            self._key,
            self._iv,
            self._ct,
            self._tag
        ])