"""Declares :class:`JSONWebKeySet`."""
import asyncio
import json
import typing

import httpx

from ckms.basekeyset import BaseKeySet
from . import models
from .exceptions import MisconfiguredIssuer


class JSONWebKeySet(BaseKeySet):
    __module__: str = 'ckms.jose'

    @staticmethod
    def http_client_factory(issuer) -> httpx.AsyncClient:
        return httpx.AsyncClient(base_url=issuer)

    @classmethod
    async def fromissuer(
        cls,
        issuer: str,
        path: typing.Optional[str] = None,
        silent: bool = False
    ) -> 'JSONWebKeySet': # pragma: no cover
        """Instantiate a new :class:`JSONWebKeySet` from an
        OAuth 2.0/OpenID Connect issuer. The URL of the issuers'
        JSON Web Key Set (JWKS) is discovered from the OAuth 2.0
        discovery endpoint.

        If no connection could be made to the specified URL or the
        JWKS content was invalid, an empty :class:`JSONWebKeySet`
        is returned.
        """
        # Attempt the OAuth 2.0 discovery URL first. If that returns
        # a 404 then try the OpenID Connect conventional metadata
        # URL.
        metadata_path = '/.well-known/oauth-authorization-server'
        if path:
            metadata_path = f"{metadata_path}/{str.lstrip(path, '/')}"
        jwks = models.JSONWebKeySet()

        # Maintain URLs that we tried so we can inform the caller
        # on failure.
        tried = []

        async with cls.http_client_factory(issuer) as session:
            try:
                response = await session.get(metadata_path)
            except httpx.ConnectError:
                if not silent:
                    raise MisconfiguredIssuer(
                        code="ISSUER_CONNECTION_ERROR",
                        http_status_code=503
                    )
            if response.status_code == 404:
                tried.append(str(response.url))
                response = await session.get(
                    url=str.replace(
                        metadata_path,
                        'oauth-authorization-server',
                        'openid-configuration'
                    )
                )
            if 200 <= response.status_code < 400:
                metadata = response.json()
                jwks_uri = metadata.get('jwks_uri')
                if jwks_uri is not None:
                    response = await session.get(jwks_uri)
                    if response.status_code == 404:
                        raise MisconfiguredIssuer(
                            code="ISSUER_JWKS_URI_NOT_FOUND",
                            http_status_code=503
                        )
                    jwks = models.JSONWebKeySet(**response.json())
            elif response.status_code == 404 and not silent:
                # Cannot lookup metadata.
                raise MisconfiguredIssuer(
                    code="ISSUER_METADATA_NOT_FOUND",
                    http_status_code=503
                )

        return await cls.fromjwks(jwks)

    @classmethod
    def fromkeychain(
        cls,
        keychain: BaseKeySet,
        keys: typing.List[str]
    ) -> 'JSONWebKeySet':
        """Construct a new :class:`JSONWebkeySet` from a :class:`BaseKeySet`
        instance.
        """
        instance = cls()
        for kid in keys:
            instance.extend(cls.to_jwk(keychain, kid))
        return instance

    @classmethod
    async def fromjwks(cls, jwks: models.JSONWebKeySet):
        """Create a new :class:`JSONWebKeySet` from an instance containing
        the input JWKS.
        """
        instance = cls()
        for jwk in jwks.keys:
            instance.add(jwk)
        await instance.setup()
        return instance

    @classmethod
    def fromtags(cls, keychain: BaseKeySet, tags: typing.List[str]) -> 'JSONWebKeySet':
        """Create a new :class:`JSONWebKeySet` from tagged keys."""
        instance = cls()
        for kid in keychain.get_tagged(tags):
            instance.extend(cls.to_jwk(keychain, kid))
        return instance

    @staticmethod
    def to_jwk(keychain: BaseKeySet, kid: str):
        """Lookup `kid` from keychain and convert it to a JSON Web Key (JWK)."""
        return keychain.get_public_key_versions(
            kid=kid,
            factory=lambda spec: models.JSONWebKey.fromimpl(**spec.dict())
        )

    def get_key_identifiers(self) -> set:
        """Return the set of key identifiers that are contained in this
        :class:`JSONWebKeySet`.
        """
        return set(self._keys.keys())

    def has_public_key(self, kid: str) -> bool:
        """Return a boolean indicating if the public key identified by `kid`
        is known by the :class:`JSONWebKeySet`.
        """
        return kid in self.get_key_identifiers()

    def dict(self) -> dict:
        """Serialize the :class:`JSONWebKeySet`."""
        obj = models.JSONWebKeySet(
            keys=list(self._keys.values())
        )
        return obj.dict(exclude_defaults=True)

    def extend(self, keys: typing.List[models.JSONWebKey]) -> None:
        """Extend the :class:`JSONWebKeySet` with the given list of
        keys.
        """
        for jwk in keys:
            assert jwk.kid is not None, jwk # nosec
            self.add(jwk)

    async def verify(
        self,
        signature: bytes,
        data: bytes,
        *,
        algorithm: str = None,
        using: str = None,
    ) -> bool:
        """Return a boolean indicating if `signature` is valid for `data."""
        verify = super().verify
        if using:
            jwk = self.get(using)
            if not algorithm and not jwk.alg: # pragma: no cover
                raise TypeError("The `algorithm` parameter is required.")
            algorithm = algorithm or jwk.alg
            return await verify(
                signature=signature,
                data=data,
                algorithm=algorithm,
                using=using
            )

        # If no kid was specified, try all keys.
        futures = []
        for jwk in self:
            if not algorithm and not jwk.alg: # pragma: no cover
                raise TypeError("The `algorithm` parameter is required.")
            algorithm = algorithm or jwk.alg
            futures.append(
                verify(
                    signature=signature,
                    data=data,
                    algorithm=algorithm,
                    using=jwk.kid
                )
            )
        return any(await asyncio.gather(*futures))

    def __len__(self):
        return len(self._keys.values())

    def __getitem__(self, index: typing.Union[int, slice]) -> models.JSONWebKey:
        return list(self._keys.items())[index][1]
