"""Declares :class:`JSONWebSignature`."""
import asyncio
import typing

from ckms import Keychain
from ckms.utils import b64decode
from ckms.utils import b64encode
from .claimset import ClaimSet
from .exceptions import MalformedToken
from .exceptions import AlgorithmNotAllowed
from .joseheader import JOSEHeader
from .josepayloadcontainer import JOSEPayloadContainer
from .jsonwebtoken import JSONWebToken
from .signature import Signature


class JSONWebSignature(JOSEPayloadContainer):
    __module__: str = 'ckms.jose'

    @property
    def typ(self) -> typing.Optional[str]:
        return self._signatures[0].typ

    @property
    def header(self):
        return self._signatures[0]._jose

    @classmethod
    def parse(cls, token: typing.Union[bytes, str]) -> 'JSONWebSignature':
        """Parse a JSON Web Signature (JWS) from JWS Compact Serialization
        and return a :class:`JSONWebSignature` instance.
        """
        if not isinstance(token, bytes): # pragma: no cover
            token = str.encode(token, "ascii")
        try:
            header, payload, signature = bytes.split(token, b'.')
        except ValueError:
            raise MalformedToken
        return JSONWebSignature(
            payload=payload,
            signatures=[Signature(
                signature=b64decode(signature),
                protected=header
            )]
        )

    @classmethod
    def frompayload(cls, payload: bytes) -> 'JSONWebSignature':
        """Return a new, unsigned :class:`JSONWebSignature` from
        the byte-sequence `payload`.
        """
        return cls(payload=b64encode(payload))

    def __init__(
        self,
        payload: bytes,
        signatures: typing.List[Signature] = None
    ):
        self._payload = payload
        self._signatures = signatures or []

    def get_claimset(self) -> ClaimSet:
        """Return a :class:`~ckms.jose.ClaimSet` instance or
        raise an error if the :class:`JSONWebSignature` does
        not contain a JWT.
        """
        return ClaimSet.frompayload(self._payload)

    async def get_payload(
        self,
        codec: 'PayloadCodec',
        verify: bool = True,
        allow_none: bool = False
    ) -> typing.Union[bytes, str, 'JSONWebToken']:
        """Return the payload encapsulated by the JSON Web Signature (JWS)."""
        if len(self._signatures) > 1:
            raise NotImplementedError
        signature = self._signatures[0]
        if signature.cty:
            # If the is a `cty` claim in the header, then a JWT is contained
            # in this container, presumable a JWE.
            raise NotImplementedError
        payload = self._payload
        if signature.algorithm == "none":
            if not allow_none:
                raise AlgorithmNotAllowed
            return codec.deserialize_jwt(payload)

        if verify:
            await codec.verify_signature(signature, payload, allow_none=allow_none)
        return codec.deserialize_jwt(payload)\
            if str.lower(signature.typ or '') in {"at+jwt", "jwt"}\
            else codec.deserialize_payload(payload)

    def is_signed(self) -> bool:
        """Return a boolean indicating if the :class:`JSONWebSignature` has
        at least one signature.
        """
        return bool(self._signatures)

    def is_encrypted(self) -> bool:
        """Return a boolean indicating if the payload is a JWE."""
        return False

    def serialize_compact(self) -> bytes:
        """Serialize the :class:`JSONWebSignature` using the JWS Compact
        Serialization format. Note that this method will raise an error
        if there are multiple signatures, since that format only supports
        a single format.
        """
        if len(self._signatures) > 1:
            raise NotImplementedError
        return b'.'.join([
            bytes(self._signatures[0].protected),
            bytes(self._payload),
            bytes(self._signatures[0].serialize())
        ])

    async def sign(
        self,
        keychain: Keychain,
        algorithm: str,
        using: str,
        claims: typing.Optional[typing.Dict[str, typing.Any]] = None,
        typ: typing.Optional[str] = None
    ) -> Signature:
        """Add a signature to the :class:`~ckms.jose.JSONWebSignature`
        using the given `signer`, key `using` and additional header
        claims `claims`.
        """
        claims = claims or {}
        claims['alg'] = algorithm
        if algorithm != "none" and 'kid' not in claims:
            meta = keychain.get_key_metadata(using)
            claims['kid'] = meta.get_signing_kid()
            assert claims.get('kid')
        if typ is not None and 'typ' not in claims:
            claims['typ'] = typ
        self._signatures.append(
            await Signature.create(
                keychain=keychain,
                algorithm=algorithm,
                using=using,
                payload=bytes(self._payload),
                protected=bytes(JOSEHeader(**claims))
            )
        )
        return self._signatures[-1]

    async def verify(
        self,
        keychain: Keychain
    ) -> bool:
        """Return a boolean indicating if all signatures are valid."""
        futures = []
        for sig in self._signatures:
            futures.append(sig.verify(keychain, self._payload))
        return all(await asyncio.gather(*futures))

    def __bytes__(self) -> bytes: # pragma: no cover
        return self.serialize_compact()
