"""Declares :class:`JSONWebToken`."""
import asyncio
import datetime
import json
import typing

from ckms import Keychain
from ckms.utils import b64decode
from ckms.utils import b64encode
from ckms.utils import current_timestamp
from . import models
from .audiencetype import AudienceType
from .claimdescriptor import ClaimDescriptor
from .exceptions import Expired
from .exceptions import InvalidClaim
from .exceptions import InvalidClaimType
from .exceptions import NotEffective
from .exceptions import WrongAudience
from .validator import Validator


class JSONWebToken:
    __module__: str = 'ckms.jose'
    _claims: dict[str, typing.Any]

    @classmethod
    def fromb64(cls, encoded: bytes) -> 'JSONWebToken':
        """Decode a :class:`JOSEHeader` instance from am
        URL-encoded, base64 byte-sequence.
        """
        return cls(**json.loads(b64decode(encoded)))

    @property
    def aud(self) -> set:
        """Return the ``aud`` claim, or an empty set."""
        claim = self._claims.get('aud')
        if claim is None: # pragma: no cover
            return set()
        if not isinstance(claim, (list, str, AudienceType)):
            raise InvalidClaimType('aud')
        if not isinstance(claim, (list, AudienceType)):
            claim = [claim]
        if not all([isinstance(x, str) for x in claim]):
            raise InvalidClaimType('aud')
        return set(sorted(claim))

    @property
    def claims(self) -> dict[str, typing.Any]:
        """The claims that were specified on the JWT."""
        return dict(self._claims)

    # RFC 7519 claims
    iss: str = ClaimDescriptor('iss', types=[str])
    jti: typing.Any = ClaimDescriptor('jti')
    sub: typing.Any = ClaimDescriptor('sub')
    exp: int = ClaimDescriptor('exp', types=[int])
    iat: int = ClaimDescriptor('iat', types=[int])
    nbf: int = ClaimDescriptor('nbf', types=[int])

    def __init__(self, **claims: typing.Any):
        self._claims = claims
        self._payload = None

    def expires_in(
        self,
        ttl: typing.Optional[int] = None,
        now: typing.Optional[int] = None
    ) -> int:
        now = now if now is not None else current_timestamp()
        if ttl is None:
            return self.exp
        self._claims['exp'] = now + ttl
        return self.exp

    async def validate(
        self,
        validators: typing.List[Validator] = None,
        now: typing.Optional[int] = None,
        audience: typing.Optional[set] = None
    ) -> None:
        """Validate the claims presented in the JSON Web Token (JWT).
        Optionally use `validator` to perform additional validation.

        Args:
            validator (:class:`ckms.jose.Validator`): a :class:`Validator`
                instance to perform additional validation. Defaults to
                ``None``, meaning that no additional validation is
                performed.
            audience (set): additional audiences of at least one must
                be specified by the ``aud`` claim.

        Raises
            :exc:`ckms.jose.InvalidClaimType`: one or more of the claims
                presented in the JWT is of an invalid type.
            :exc:`ckms.jose.Expired`: based on the value in the ``exp``
                claim, the token is expired.
            :exc:`ckms.jose.NotEffective`: based on the value in the ``nbf``
                claim, the token is used before its effective date.

        Returns:
            None
        """
        now = now or current_timestamp()
        self.validate_claims(now, audience=audience)
        await asyncio.gather(*[
            v.validate(self, now)
            for v in (validators or []) if v
        ])

    def validate_claims(self,
        now: typing.Optional[int] = None,
        audience: typing.Optional[set] = None
    ):
        now = now or current_timestamp()
        self.validate_exp(now=now)
        self.validate_iat(now=now)
        self.validate_nbf(now=now)
        self.validate_audience(audience=audience)

    def validate_audience(self, audience: typing.Optional[str] = None):
        audience = set(audience or [])
        if (self.aud or audience) and not bool(self.aud & audience):
            raise WrongAudience(audience)

    def validate_exp(self, now: int):
        if self.exp is not None and self.exp <= now:
            raise Expired(now)

    def validate_iat(self, now: int):
        if self.iat is not None and self.iat > now:
            raise InvalidClaim(
                claim="iat",
                detail=(
                    f"The `iat` claim can not be in the future. {self.iat - now} "
                    "seconds remain until the token is effective."
                )
            )

    def validate_nbf(self, now: int):
        if self.nbf is not None and self.nbf > now:
            raise NotEffective(now)

    def get(self, claim: str) -> typing.Union[dict, int, list, str, None]:
        """Lookup the claim `claim`."""
        return self._claims.get(claim)

    def __bytes__(self) -> bytes:
        if len(self.aud) == 1: # TODO
            self._claims['aud'] = self.aud.pop()
        return b64encode(
            str.encode(json.dumps(self._claims, sort_keys=True), "utf-8")
        )

    def __iter__(self):
        return iter(self.claims.items())