"""Declares :class:`JSONWebKey`."""
import typing
import os

import pydantic
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import ed448
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import x448
from cryptography.hazmat.primitives.asymmetric import x25519
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.kdf.concatkdf import ConcatKDFHash

from ckms.algorithms import Algorithm
from ckms.basekeyset import BaseKeySet
from ckms.providers import Provider
from ckms.providers.local import Cipher
from ckms.utils import b64encode
from ckms.utils import b64encode_int
from ckms.utils import b64decode
from ckms.utils import b64decode_int
from ..exceptions import UnsupportedAlgorithm
from .const import CURVE_MAPPING
from .const import KTY_ENC_ALG_ALGORITHMS
from .const import KTY_SIG_ALGORITHMS
from .const import OKP_KEY_CLASSES


__all__ = ['JSONWebKey']


KeyType = typing.Union[
    ec.EllipticCurvePublicKey,
    ec.EllipticCurvePrivateKey,
    ed448.Ed448PublicKey,
    ed448.Ed448PrivateKey,
    ed25519.Ed25519PublicKey,
    ed25519.Ed25519PrivateKey,
    rsa.RSAPublicKey,
    rsa.RSAPrivateKey,
    bytes
]


class JSONWebKey(pydantic.BaseModel):
    __module__: str = 'ckms.jose.models'

    kty: str
    kid: typing.Optional[str]
    alg: typing.Optional[str]
    crv: typing.Optional[str]
    use: typing.Optional[str]
    key_ops: typing.Optional[typing.List[str]] = []
    x5u: typing.Optional[str]
    x5c: typing.Optional[str]
    x5t: typing.Optional[str]
    x5t_sha256: typing.Optional[str] = pydantic.Field(
        default=None,
        alias="x5t#S256"
    )

    # Key material
    d: typing.Optional[str]
    e: typing.Optional[str]
    n: typing.Optional[str]
    k: typing.Optional[str]
    p: typing.Optional[str]
    q: typing.Optional[str]
    x: typing.Optional[str]
    y: typing.Optional[str]
    dp: typing.Optional[str]
    dq: typing.Optional[str]
    qi: typing.Optional[str]

    @property
    def key(self) -> KeyType:
        return self.get_impl()

    @property
    def provider(self) -> Provider:
        return BaseKeySet.provider('local')

    @property
    def public(self) -> 'JSONWebKey':
        return JSONWebKey.fromimpl(
            typing.cast(KeyType, self.key.public_key()),
            alg=self.alg,
            kid=self.kid,
            use=self.use,
            key_ops=self.key_ops
        )

    @classmethod
    def frompem(
        cls,
        pem: typing.Union[bytes, str],
        **claims: typing.Dict[str, typing.Any]
    ) -> 'JSONWebKey':
        """Instantiate a new :class:`JSONWebKey` from a PEM-encoded public or
        private key.
        """
        if isinstance(pem, str):
            pem = str.encode(pem, 'ascii')
        header = bytes.split(pem, b'\n')[0]
        if header == b'-----BEGIN PUBLIC KEY-----':
            f = serialization.load_pem_public_key
        elif header == b'-----BEGIN PRIVATE KEY-----':
            f = lambda x: serialization.load_pem_private_key(x, None)
        else:
            raise NotImplementedError
        return cls.fromimpl(f(pem), **claims)

    @classmethod
    def fromimpl(
        cls,
        key: KeyType,
        **claims: typing.Dict[str, typing.Any]
    ) -> 'JSONWebKey':
        """Create a :class:`JSONWebKey` from a concrete implementation."""
        if isinstance(key, rsa.RSAPublicKey):
            numbers = key.public_numbers()
            claims.update({
                'kty': "RSA",
                'n': b64encode_int(numbers.n),
                'e': b64encode_int(numbers.e)
            })

        elif isinstance(key, ec.EllipticCurvePublicKey):
            numbers = key.public_numbers()
            claims.update({
                'kty': "EC",
                'crv': CURVE_MAPPING[key.curve.name],
                'x': b64encode_int(numbers.x),
                'y': b64encode_int(numbers.y)
            })
        elif isinstance(key, ed448.Ed448PublicKey):
            claims.update({
                'kty': "OKP",
                'crv': "Ed448",
                'x': b64encode(
                    key.public_bytes(
                        encoding=serialization.Encoding.Raw,
                        format=serialization.PublicFormat.Raw
                    )
                )
            })
        elif isinstance(key, x448.X448PublicKey):
            claims.update({
                'kty': "OKP",
                'crv': "X448",
                'x': b64encode(
                    key.public_bytes(
                        encoding=serialization.Encoding.Raw,
                        format=serialization.PublicFormat.Raw
                    )
                )
            })
        elif isinstance(key, ed25519.Ed25519PublicKey):
            claims.update({
                'kty': "OKP",
                'crv': "Ed25519",
                'x': b64encode(
                    key.public_bytes(
                        encoding=serialization.Encoding.Raw,
                        format=serialization.PublicFormat.Raw
                    )
                )
            })
        elif isinstance(key, x25519.X25519PublicKey):
            claims.update({
                'kty': "OKP",
                'crv': "X25519",
                'x': b64encode(
                    key.public_bytes(
                        encoding=serialization.Encoding.Raw,
                        format=serialization.PublicFormat.Raw
                    )
                )
            })
        elif isinstance(key, rsa.RSAPrivateKey):
            numbers = key.private_numbers()
            claims.update({
                'kty': "RSA",
                'n': b64encode_int(numbers.public_numbers.n),
                'e': b64encode_int(numbers.public_numbers.e),
                'd': b64encode_int(numbers.d),
                'p': b64encode_int(numbers.p),
                'q': b64encode_int(numbers.q),
                'dp': b64encode_int(numbers.dmp1),
                'dq': b64encode_int(numbers.dmq1),
                'qi': b64encode_int(numbers.iqmp),
            })
        elif isinstance(key, ec.EllipticCurvePrivateKey):
            public_numbers = key.public_key().public_numbers()
            numbers = key.private_numbers()
            claims.update({
                'kty': "EC",
                'crv': CURVE_MAPPING[key.curve.name],
                'x': b64encode_int(public_numbers.x),
                'y': b64encode_int(public_numbers.y),
                'd': b64encode_int(numbers.private_value)
            })
        elif isinstance(key, tuple(OKP_KEY_CLASSES.values())):
            public = key.public_key()
            claims.update({
                'kty': "OKP",
                'x': b64encode(
                    public.public_bytes(
                        encoding=serialization.Encoding.Raw,
                        format=serialization.PublicFormat.Raw
                    )
                ),
                'd': b64encode(
                    key.private_bytes(
                        encoding=serialization.Encoding.Raw,
                        format=serialization.PrivateFormat.Raw,
                        encryption_algorithm=serialization.NoEncryption()
                    )
                )
            })
        elif isinstance(key, Cipher):
            claims.update({
                'kty': "oct",
                'k': b64encode(bytes(key))
            })
        elif isinstance(key, bytes):
            claims.update({
                'kty': "oct",
                'k': b64encode(key)
            })
        else:
            raise NotImplementedError(repr(key))

        # Set more parameters based on key type.
        # The key type used with these keys is "OKP" and the algorithm used for
        # signing is "EdDSA".  These subtypes MUST NOT be used for Elliptic
        # Curve Diffie-Hellman Ephemeral Static (ECDH-ES).
        if isinstance(key, (ed448.Ed448PrivateKey, ed448.Ed448PublicKey)):
            claims['crv'] = "Ed448"
            claims['alg'] = 'EdDSA'
            claims['use'] = 'sig'
        elif isinstance(key, (ed25519.Ed25519PrivateKey, ed25519.Ed25519PublicKey)):
            claims['crv'] = "Ed25519"
            claims['alg'] = 'EdDSA'
            claims['use'] = 'sig'
        elif isinstance(key, (x448.X448PrivateKey, x448.X448PublicKey)):
            claims['crv'] = "X448"
            claims['use'] = 'enc'
            claims['alg'] = "ECDH-ES+A256KW"
        elif isinstance(key, (x25519.X25519PrivateKey, x25519.X25519PublicKey)):
            claims['crv'] = "X25519"
            claims['use'] = 'enc'
            claims['alg'] = "ECDH-ES+A256KW"

        return cls(**claims)

    @classmethod
    def generate(
        cls,
        kty: str,
        crv: typing.Optional[str] = None,
        length: typing.Optional[int] = None,
        **claims: typing.Dict[str, typing.Any]
    ) -> 'JSONWebKey':
        """Generate a new key using specified parameters."""
        k = None
        if kty in {"RSA", "oct"} and not isinstance(length, int):
            raise TypeError(
                "The `length` parameter must be a positive integer"
            )
        if kty in {"EC"} and crv not in CURVE_MAPPING:
            raise ValueError(f"Unknown curve: {crv}")
        if kty == "RSA":
            assert length is not None # nosec
            k =rsa.generate_private_key(65537, length)
        elif kty == "EC":
            k = ec.generate_private_key(curve=CURVE_MAPPING[crv]()) # type: ignore
        elif kty == "OKP":
            assert crv in {"Ed448", "Ed25519", "X448", "X25519"} # nosec
            key_class = OKP_KEY_CLASSES[crv]
            k = key_class.generate()
        elif kty == "oct":
            assert length is not None # nosec
            k = os.urandom(length)
        assert k is not None # nosec
        return cls.fromimpl(k, **claims)

    def __init__(self, **data: typing.Any) -> None:
        super().__init__(**data)

    def exchange(self, peer: 'JSONWebKey', kdf: ConcatKDFHash) -> bytes:
        """Perform a key exchange with the peer public key."""
        sk = self.key.exchange(peer.key)
        return (
            JSONWebKey.fromimpl(sk),
            JSONWebKey.fromimpl(kdf.derive(sk))
        )

    def is_loaded(self) -> bool: # pragma: no cover
        return True

    def get_impl(
        self
    ) -> KeyType:
        """Return the concrete key implemetation described by the parameters
        of the JSON Web Key (JWK).
        """
        if self.kty == "RSA":
            assert self.n # nosec
            assert self.e # nosec
            numbers = rsa.RSAPublicNumbers(
                n=b64decode_int(self.n),
                e=b64decode_int(self.e),
            )
            key = numbers.public_key()
            if None not in {self.d, self.p, self.q, self.dp, self.dq, self.qi}:
                assert self.d # nosec
                assert self.p # nosec
                assert self.q # nosec
                assert self.dp # nosec
                assert self.dq # nosec
                assert self.qi # nosec
                numbers = rsa.RSAPrivateNumbers(
                    public_numbers=numbers,
                    d=b64decode_int(self.d),
                    p=b64decode_int(self.p),
                    q=b64decode_int(self.q),
                    dmp1=b64decode_int(self.dp),
                    dmq1=b64decode_int(self.dq),
                    iqmp=b64decode_int(self.qi)
                )
                key = numbers.private_key()
        elif self.kty == "EC":
            assert self.x and self.y # nosec
            numbers = ec.EllipticCurvePublicNumbers(
                x=b64decode_int(self.x),
                y=b64decode_int(self.y),
                curve=CURVE_MAPPING[self.crv]()
            )
            key = numbers.public_key()
            if self.d is not None:
                numbers = ec.EllipticCurvePrivateNumbers(
                    private_value=b64decode_int(self.d),
                    public_numbers=numbers
                )
                key = numbers.private_key()
        elif self.kty == "OKP" and self.crv == "Ed448":
            assert self.x is not None # nosec
            key = ed448.Ed448PublicKey.from_public_bytes(b64decode(self.x))
            if self.d is not None:
                key = ed448.Ed448PrivateKey.from_private_bytes(b64decode(self.d))
        elif self.kty == "OKP" and self.crv == "X448":
            assert self.x is not None # nosec
            key = x448.X448PublicKey.from_public_bytes(b64decode(self.x))
            if self.d is not None:
                key = x448.X448PrivateKey.from_private_bytes(b64decode(self.d))
        elif self.kty == "OKP" and self.crv == "Ed25519":
            assert self.x is not None # nosec
            key = ed25519.Ed25519PublicKey.from_public_bytes(b64decode(self.x))
            if self.d is not None:
                key = ed25519.Ed25519PrivateKey.from_private_bytes(b64decode(self.d))
        elif self.kty == "OKP" and self.crv == "X25519":
            assert self.x is not None # nosec
            key = x25519.X25519PublicKey.from_public_bytes(b64decode(self.x))
            if self.d is not None:
                key = x25519.X25519PrivateKey.from_private_bytes(b64decode(self.d))
        elif self.kty == "oct": # pragma: no cover
            assert self.k is not None # nosec
            key = b64decode(self.k)
        else: # pragma: no cover
            raise NotImplementedError
        return key

    def decrypt(self, algorithm: str, data: bytes) -> bytes:
        """Encrypt `data` using the :class:`JSONWebKey` using `algorithm`."""
        alg = Algorithm.get(algorithm)
        return alg.decrypt(self.provider, self, data) # type: ignore

    def encrypt(self, algorithm: str, data: bytes) -> bytes:
        """Encrypt `data` using the :class:`JSONWebKey` using `algorithm`."""
        alg = Algorithm.get(algorithm)
        return alg.encrypt(self.provider, self, data) # type: ignore

    def sign(self, algorithm: str, data: bytes) -> bytes:
        """Sign `data` using the :class:`JSONWebKey` using `algorithm`."""
        alg = Algorithm.get(algorithm)
        return alg.sign(self.provider, self, data) # type: ignore

    def verify(
        self,
        algorithm: str,
        signature: bytes,
        data: bytes
    ) -> bytes:
        """Verify that this :class:`JSONWebKey` was used to create
        `signature` from `data`.
        """
        if not self.supports_signing_algorithm(algorithm):
            raise UnsupportedAlgorithm(
                given=algorithm,
                supported=self.get_signing_algorithms()
            )
        alg = Algorithm.get(algorithm)
        return alg.verify( # type: ignore
            provider=self.provider,
            key=self,
            signature=signature,
            data=data
        )

    def get_encryption_algorithms(self) -> typing.Set[str]:
        """Return the set of encryption algorithms supported by the JWK."""
        if self.alg is not None:
            return {self.alg}

        crv = self.crv if self.kty == "EC" else None
        if (self.kty, crv) not in KTY_ENC_ALG_ALGORITHMS:
            return set()
        return KTY_ENC_ALG_ALGORITHMS[(self.kty, crv)]

    def get_signing_algorithms(self) -> typing.Set[str]:
        """Return the set of signing algorithms supported by the JWK."""
        if self.alg is not None:
            return {self.alg}

        crv = self.crv if self.kty == "EC" else None
        if (self.kty, crv) not in KTY_SIG_ALGORITHMS:
            return set()
        return KTY_SIG_ALGORITHMS[(self.kty, crv)]

    def supports_signing_algorithm(self, algorithm: str) -> bool:
        """Return a boolean indicating if the key supports the given
        signing algorithm.
        """
        return algorithm in self.get_signing_algorithms()
