"""Declares :class:`JSONWebKey`."""
import asyncio
import itertools
import typing

import httpx
import pydantic

from ckms import BaseKeySet
from ckms import Keychain
from ..jsonwebsignature import JSONWebSignature
from ..signature import Signature
from .jsonwebkey import JSONWebKey


class JSONWebKeySet(pydantic.BaseModel):
    __module__: str = 'ckms.jose.models'
    keys: typing.List[JSONWebKey] = []

    @classmethod
    async def discover(cls, issuer: str) -> 'JSONWebKeySet':
        """Instantiate a new :class:`JSONWebKeySet` from the OAuth 2.0
        Server Metadata or OpenID Configuration endpoints.
        """
        urls = [
            'oauth-authorization-server',
            'openid-configuration'
        ]
        jwks = cls()
        async with httpx.AsyncClient(base_url=issuer) as client:
            try:
                metadata = None
                for path in urls:
                    response = await client.get(f'/.well-known/{path}')
                    if response.status_code == 200:
                        metadata = response.json()
                        break
                if metadata is not None and metadata.get('jwks_uri'):
                    response = await client.get(metadata['jwks_uri'])
                    response.raise_for_status()
                    jwks = cls(**response.json())
            except (httpx.HTTPError, httpx.StreamError, TypeError, ValueError):
                raise
        return jwks

    @classmethod
    def fromtags(
        cls,
        keyset: BaseKeySet,
        tags: typing.List[str],
        force_public: bool = True
    ) -> 'JSONWebKeySet':
        """Lookup keys from `keyset` tagged with any of `tags` and
        return a new :class:`JSONWebKetSet` holding the keys.
        """
        if not force_public: # pragma: no cover
            raise NotImplementedError
        keys: typing.List[JSONWebKey] = []
        for kid in keyset.get_tagged(tags):
            jwk: typing.List[JSONWebKey] = keyset.get_public_key_versions(
                kid=kid,
                factory=lambda spec: JSONWebKey.fromimpl(**spec.dict()) # type: ignore
            )
            keys.extend(jwk)
        return cls(keys=keys)


    @classmethod
    def fromkeychain(
        cls,
        keyset: BaseKeySet | Keychain,
        force_public: bool = True
    ) -> 'JSONWebKeySet':
        """Convert a :class:`ckms.core.BaseKeySet` to a JWKS."""
        if not force_public: # pragma: no cover
            raise NotImplementedError
        keys: typing.List[JSONWebKey] = []
        for kid in keyset._keys:
            jwk: typing.List[JSONWebKey] = keyset.get_public_key_versions(
                kid=kid,
                factory=lambda spec: JSONWebKey.fromimpl(**spec.dict()) # type: ignore
            )
            keys.extend(jwk)
        return cls(keys=keys)

    def get_encryption_key(self) -> JSONWebKey | None:
        """Return the preferred encryption key."""
        keys = self.get_encryption_keys()
        return keys[0] if keys else None

    def get_encryption_keys(self) -> list[JSONWebKey]:
        """Return all keys that may be used for encryption or
        decryption.
        """
        return [jwk for jwk in self.keys if jwk.use == "enc"]

    def get_encryption_algorithms(self) -> typing.List[str]:
        """Return the list of supported encryption algorithms of this
        :class:`JSONWebKeySet`.
        """
        algorithms = itertools.chain(*[
            x.get_encryption_algorithms() for x in self.keys if x.use == 'enc']
        )
        return list(sorted(set(algorithms)))

    def get_signing_keys(self) -> typing.List[JSONWebKey]:
        """Return all keys that may be used for signing or
        signature verification.
        """
        return [jwk for jwk in self.keys if jwk.use == "sig"]

    def get_signing_algorithms(self) -> typing.List[str]:
        """Return the list of supported signature algorithms of this
        :class:`JSONWebKeySet`.
        """
        algorithms = itertools.chain(*[
            x.get_signing_algorithms() for x in self.keys if x.use == 'sig']
        )
        return list(sorted(set(algorithms)))

    def has_key(self, kid: str) -> bool:
        """Return a boolean indicating if the given key is in the
        :class:`JSONWebKeySet`.
        """
        for key in self.keys:
            if key.kid != kid:
                continue
            result = True
            break
        else:
            result = False
        return result

    def verify(
        self,
        signature: typing.Union[
            bytes,
            JSONWebSignature,
            Signature
        ],
        data: typing.Optional[bytes] = None,
        using: typing.Optional[str] = None,
        algorithm: typing.Optional[str] = None
    ):
        """Verify a :class:`~ckms.jose.JSONWebSignature` instance
        using the keys in the :class:`JSONWebKeySet`.
        """
        result = None
        if isinstance(signature, JSONWebSignature):
            result = signature.verify(self)
        elif isinstance(signature, Signature):
            result = signature.verify(
                keychain=self,
                payload=data,
                using=using
            )
        elif isinstance(signature, bytes):
            if data is None:
                raise TypeError("The `data` parameter must be a byte-sequence.")
            if algorithm is None:
                raise TypeError("The `algorithm` parameter can not be None.")
            futures = []
            for key in self.keys:
                if key.use not in {"sig", None}\
                or (using is not None and key.kid != using)\
                or not key.supports_signing_algorithm(algorithm):
                    continue
                futures.append(key.verify(algorithm, signature, data))
            result = self.any(futures)
        else:
            raise NotImplementedError(type(signature))
        return result or self.false()

    async def false(self) -> bool: # pragma: no cover
        return False

    async def any(self, futures: typing.List[typing.Any]):
        return any(await asyncio.gather(*futures))

    def __len__(self) -> int:
        return len(self.keys)