"""Declares :class:`PayloadCodec`."""
import functools
import typing

import ckms
from ckms import ops
from ckms import providers
from ckms.fixedalgorithmkey import FixedAlgorithmKey
from ckms.basekeyset import BaseKeySet
from ckms.utils import b64decode
from ckms.utils import b64encode
from .claimset import ClaimSet
from .exceptions import AlgorithmNotAllowed
from .exceptions import InvalidSignature
from .exceptions import InvalidHeader
from .exceptions import UnsupportedContentType
from .exceptions import UnsupportedTokenType
from .exceptions import Unverifiable
from .josedeserializer import JOSEDeserializer
from .joseheader import JOSEHeader
from .jsonwebencryption import JSONWebEncryption
from .jsonwebsignature import JSONWebSignature
from .jsonwebtoken import JSONWebToken
from .models import JSONWebKeySet
from .payload import Payload
from .signature import Signature
from .validator import Validator


SUPPORTED_ALGORITHMS = [
    "RSA1_5",
    "RSA-OAEP",
    "RSA-OAEP-256",
    "RSA-OAEP-384",
    "RSA-OAEP-512",
    "A128KW",
    "A192KW",
    "A256KW",
    "A128GCMKW",
    "A192GCMKW",
    "A256GCMKW",
    "dir",
    "ECDH-ES",
    "ECDH-ES+A128KW",
    "ECDH-ES+A192KW",
    "ECDH-ES+A256KW",
    "PBES2-HS256+A128KW",
    "PBES2-HS256+A192KW",
    "PBES2-HS256+A256KW",
]


class PayloadCodec:
    """Exposes an interface to decode and encode a JSON Web Token (JWT),
    JSON Web Encryption (JWE) or JSON Web Signature (JWS).
    """
    __module__: str = 'ckms.jose'
    signer: BaseKeySet
    verifier: BaseKeySet
    decrypter: BaseKeySet
    encrypter: BaseKeySet
    serializer: JOSEDeserializer
    allow_jwks_import: bool
    trusted_issuers: set
    validator: Validator

    def __init__(
        self,
        signer: BaseKeySet | JSONWebKeySet | None = None,
        verifier: BaseKeySet | JSONWebKeySet | None = None,
        decrypter: BaseKeySet | JSONWebKeySet | None = None,
        encrypter: BaseKeySet | JSONWebKeySet | None = None,
        random: typing.Optional[providers.Provider] = None,
        required_claims: typing.Optional[typing.Set[str]] = None,
        serializer: JOSEDeserializer = JOSEDeserializer(),
        jwks_import: bool = False,
        audiences: typing.Optional[typing.Set[str]] = None,
        trusted_issuers: typing.Optional[typing.Set[str]] = None,
        content_encryption: str = "A256GCM"
    ):
        """A :class:`PayloadCodec` instance governs the rules under which
        objects are deserialized, decoded and instantiated into the
        corresponding Python objects.

        The `verifier` and `decrypter` parameters are required and specify
        which :class:`~ckms.BaseKeySet` implementations are used to verify
        digital signatures and decrypt JWE payloads. A common configuration is
        to provide a :class:`~ckms.jose.JSONWebKeySet` holding public keys
        exclusively as the `verifier` parameter, and a :class:`ckms.Keychain`
        or similar as the `decrypter` parameter. If any of these parameters
        is omitted, then the respective operations can not be performed by
        :class:`PayloadCodec` and will if the (public) keys can not be
        resolved otherwise (e.g. through the issuer JWKS URI for example).

        A provider of entropy may be specified using the `random` argument.

        The `required_claims` specifies the claims the must be present when
        the payload is a JWT. If any of these claims is missing, then
        validation of the JWT fails. Null values are not considered missing
        claims.

        The `jwks_import` parameter is a boolean indicating if a JSON
        Web Key Set (JWKS) or JSON Web Key (JWK) may be imported using either
        the ``iss`` (using OAuth 2.0 metadata discovery) or ``jku`` claims,
        respectively. The default is ``False``.

        With the `trusted_issuers` parameter the set of issuers that are
        trusted by the codec may be specified. Any token that has an ``iss``
        claim that is not in this set, is rejected. If `jwks_import``
        is ``True``, then a JWK(S) may be imported if the issuer is in
        this set.  If `trusted_issuers` is ``None`` or an empty set, then the
        `iss` claim is not considered during JWT validation.

        Similarly, the `audiences` parameter specifies the set of audiences
        with which the set of audiences in the `aud` claim of a JWT must
        intersect. If `audiences` is ``None`` or an empty set, then the
        `aud` claim is not considered during JWT validation.
        """
        self.signer = signer or ckms.get_default_keychain()
        self.verifier = verifier or ckms.get_default_keychain()
        self.decrypter = decrypter or ckms.get_default_keychain()
        self.encrypter = encrypter or ckms.get_default_keychain()
        self.serializer = serializer
        self.jwks_import = jwks_import
        self.trusted_issuers = str(trusted_issuers or [])
        self.issuers = {}
        self.validator = None
        self.content_encryption = content_encryption
        self.random = random

    async def decode(
        self,
        token: bytes | dict[str, typing.Any],
        validator: typing.Optional[Validator] = None,
        verify: bool = True,
        validate: bool = True,
        allow_none: bool = False,
        _type: type = JSONWebToken
    ) -> typing.Union[bytes, str, JSONWebToken]:
        """Decode `token` and return the payload, which is either a
        :class:`~ckms.jose.JSONWebToken` or byte-sequence.

        The `token` parameter is a byte-sequence holding either a
        JSON Web Signature (JWS) or JSON Web Encryption (JWE), or
        a Nested JWT, or a JWS/JWE with a binary payload.

        If the expected payload is a JWT, then optionally provide
        a :class:`~ckms.jose.Validator` instance `validator` to
        perform additional validation on the claims contained in
        the JWT.

        The `verify` parameter indicates if the signature must be
        verified if the payload is a JWS. It defaults to ``True``.

        Likewise, the `validate` parameter indicates if the claims of
        a JWT must be verified.

        Raises:
            :exc:`ckms.jose.JOSEException`: generic exception
                base class that may be used to catch all of the below
                exception types.
            :exc:`ckms.jose.InvalidSignature`: the signature
                did not validate.
            :exc:`ckms.jose.InvalidClaimType`: one or more of the claims
                presented in the JWT is of an invalid type.
            :exc:`ckms.jose.Expired`: based on the value in the ``exp``
                claim, the token is expired.
            :exc:`ckms.jose.NotEffective`: based on the value in the ``nbf``
                claim, the token is used before its effective date.
        """
        if isinstance(token, dict):
            raise NotImplementedError
        payload = self.serializer.deserialize_compact(token)
        payload.accept()
        while not isinstance(payload, (bytes, str, _type)):
            payload = await payload.get_payload(
                codec=self,
                verify=verify,
                allow_none=allow_none
            )
        if isinstance(payload, JSONWebToken) and validate:
            # Run validation based on the parameters passed to the
            # constructor and the call arguments.
            await payload.validate([validator, await self.get_validator(payload)])
        return payload

    async def decode_jws(
        self,
        token: typing.Union[bytes, str],
    ) -> JSONWebSignature:
        """Like :meth:`decode()`, but returns an **unverified**
        :class:`~ckms.jose.JSONWebSignature` instance.
        """
        return await self.decode(token, verify=False, _type=JSONWebSignature)

    async def decrypt_cek(
        self,
        header: JOSEHeader,
        ct: ops.Ciphertext
    ) -> bytes:
        """Decrypt a Content Encryption Key (CEK) that was parsed from a
        JSON Web Encryption (JWE) instance.
        """
        assert isinstance(ct, ops.Ciphertext) # nosec
        if header.kid is None: # pragma: no cover
            raise InvalidHeader(
                message=(
                    "The supplied JWE can not be decrypted because the `kid` "
                    "claim is missing from the header, thus the server could "
                    "not resolve the encryption key."
                )
            )
        alg = header.alg
        if alg == 'dir':
            spec = self.decrypter.get(header.kid)
        else:
            spec = await self.decrypter.load(
                provider_name='local',
                kid=None,
                spec={
                    'secret': await self.decrypter.decrypt(
                        algorithm=alg,
                        using=header.kid,
                        data=ct
                    ),
                    'use': "enc"
                }
            )
        return FixedAlgorithmKey(algorithm=header.enc, spec=spec)

    def deserialize_jwt(self, payload: bytes) -> JSONWebToken:
        """Deserialize the payload of a JSON Web Signature (JWS) or
        JSON Web Encryption (JWE) to a JSON Web Token (JWT).

        The `payload` parameter is the undecoded payload of the JWS.
        """
        return JSONWebToken.fromb64(payload)

    def deserialize_payload(self, payload: bytes) -> bytes:
        """Deserialize the payload of a JWE or JWS."""
        return b64decode(payload)

    def encode(
        self,
        payload: bytes | dict[str, typing.Any] | ClaimSet | JSONWebToken | Payload,
        claims: dict[str, typing.Any] | None = None
    ) -> Payload:
        """Encode `payload`, which is either a byte-sequence or a
        :class:`~ckms.jose.JSONWebToken` instance.
        """
        if not isinstance(payload, Payload):
            if isinstance(payload, ClaimSet):
                payload = payload.dict(exclude_defaults=True)
            if isinstance(payload, dict):
                payload = JSONWebToken(**payload)
            assert isinstance(payload, (bytes, JSONWebToken)), type(payload) # nosec
            payload = Payload(self, payload, claims)
        return payload

    async def encrypt(
        self,
        algorithm: str,
        using: str,
        payload: Payload,
        encryption: typing.Optional[str] = None,
        claims: typing.Optional[dict] = None
    ) -> JSONWebEncryption:
        """Encrypt `payload` using the given parameters."""
        if algorithm not in SUPPORTED_ALGORITHMS:
            raise NotImplementedError(algorithm)
        encryption = encryption or self.content_encryption
        claims = claims or {}
        claims.update({
            'enc': encryption
        })
        if algorithm != 'dir':
            cek =  await FixedAlgorithmKey.generate(encryption, self.random or ckms)
            encrypt_cek = functools.partial(
                self.encrypt_cek,
                algorithm=algorithm,
                cek=cek
            )
        else:
            encrypt_cek = functools.partial(
                self.encrypt_dir,
                algorithm=encryption,
            )
            cek = FixedAlgorithmKey(
                algorithm=encryption,
                spec=self.encrypter.get(using)
            )
        header, encrypted_cek = await encrypt_cek(
            using=using,
            claims=claims
        )
        ct = await cek.encrypt(payload, aad=bytes(header))

        # A*GCMKW mode
        if hasattr(encrypted_cek, 'tag'):
            header.tag = b64encode(encrypted_cek.tag) or None
        if hasattr(encrypted_cek, 'iv'):
            header.iv = b64encode(encrypted_cek.iv) or None
        return JSONWebEncryption(
            protected=bytes(header),
            key=b64encode(bytes(encrypted_cek)),
            iv=b64encode(ct.iv),
            ct=b64encode(bytes(ct)),
            tag=b64encode(ct.tag)
        )

    async def encrypt_cek(
        self,
        algorithm: str,
        using: str,
        cek: FixedAlgorithmKey,
        claims: dict
    ) -> typing.Tuple[JOSEHeader, ops.Ciphertext]:
        """Encrypt the Content Encryption Key (CEK) with the given `algorithm`
        and key `using`. Return a tuple containing the :class:`~ckms.jose.JOSEHeader`
        and the resulting :class:`~ckms.ops.Ciphertext`.
        """
        meta = self.encrypter.get_key_metadata(using)
        params = {
            'algorithm': algorithm,
            'using': using,
            'data': bytes(cek)
        }
        claims.update({
            'alg': algorithm,
            'kid': meta.get_signing_kid()
        })
        if algorithm in {"A128GCMKW", "A192GCMKW", "A256GCMKW"}:
            # TODO
            params['data'] = ops.Plaintext(
                bytes(cek),
                aad=b''
            )
        
        assert claims.get('kid') is not None # nosec
        ct = await self.encrypter.encrypt(**params)
        return (JOSEHeader.fromcek(ct, claims), ct)

    async def encrypt_dir(
        self,
        algorithm: str,
        using: str,
        claims: dict
    ) -> typing.Tuple[JOSEHeader, bytes]:
        """Return a tuple containing a header and an empty byte-sequence, for
        use with the ``dir`` encryption algorithm.
        """
        meta = self.encrypter.get_key_metadata(using)
        claims.update({
            'alg': 'dir',
            'kid': meta.get_signing_kid(),
            'enc': algorithm
        })
        return JOSEHeader(**claims), b''

    async def get_validator(self, jwt: JSONWebToken) -> Validator:
        """Lookup a validator for a specific :class:`~ckms.jose.JSONWebToken`
        instance. The default implementation return the `validator` parameter
        that was passed to the constructor of :class:`PayloadCodec`.

        Subclasses may override this method to implement validation logic
        that is specific to the given `jwt`, for example populating a
        :class:`~ckms.jose.Validator` with a predefined list of allowed ``sub``
        values that are fetched from a database.
        """
        return self.validator

    @staticmethod
    def introspect(
        token: str | bytes | Payload,
        model: type[ClaimSet] = ClaimSet,
        accept: set[str] = {"at+jwt", "jwt"}
    ) -> tuple[JOSEHeader, ClaimSet | None]:
        """Introspect the token. Return a :class:`ckms.jose.models.JOSEHeader`
        instance, and a :class:`ckms.jose.ClaimSet` instance representing the
        claims if the token is a JSON Web Token (JWT), otherwise the second
        variable in the tuple is ``None``.

        Optionally the `model` parameter may be used to specify the model to
        parse the claims. It must be a subclass of :class:`ckms.jose.ClaimSet`.
        """
        if isinstance(token, Payload):
            token = bytes(token)
        if not isinstance(token, bytes):
            token = str.encode(token , 'ascii')
        header, *parts = token.split(b'.')
        header = JOSEHeader.fromb64(header)
        claims = None
        if str.lower(header.typ or '') in accept and not header.is_encrypted():
            payload, _ = parts
            claims = model.fromb64(payload)
        return header, claims

    def get_key_algorithm(self, using: str) -> str | None:
        """Return the algorithm for the specified key `using`."""
        meta = self.signer.get_key_metadata(using)
        return meta.alg

    async def sign_jws(
        self,
        algorithm: str,
        using: str,
        jws: JSONWebSignature,
        claims: typing.Optional[typing.Dict[str, typing.Any]] = None
    ) -> None:
        """Sign a :class:`JSONWebSignarure`."""
        await jws.sign(
            keychain=self.signer,
            algorithm=algorithm,
            using=using,
            claims=claims
        )

    async def verify_signature(
        self,
        signature: Signature,
        payload: bytes,
        allow_none: bool = False
    ):
        """Verifies a :class:`~ckms.jose.Signature` instance, attached to
        a JSON Web Signature (JWS). Raise an exception if the signature
        did not validate.

        Args:
            signature (:class:`~ckms.jose.Signature`): the signature that
                must be validated.
            payload (bytes): the payload over which the signature was
                calculated. This should be the unmodified content that
                was parsed from a token i.e. base64, URL-encoded.
            allow_none (bool): indicates if the ``none`` algorithm is
                allowed. Default is ``False``.

        Raises:
            :exc:`ckms.jose.exceptions.InvalidSignature`: the signature
                did not validate.
        """
        if signature.algorithm == "none" and not allow_none:
            raise AlgorithmNotAllowed

        # Bail out early if we don't know the key.
        if not self.verifier.has_public_key(signature.kid):
            raise Unverifiable(signature.kid)

        if not await signature.verify(self.verifier, payload):
            raise InvalidSignature
