"""Declares :class:`Signature`."""
import json

from ckms import Keychain
from ckms.utils import b64decode
from ckms.utils import b64encode
from .exceptions import AlgorithmNotAllowed
from .exceptions import MalformedToken
from .joseheader import JOSEHeader


class Signature:
    """Represents a signature on a JSON Web Signature (JWS)."""
    __module__: str = 'ckms.jose'

    @property
    def algorithm(self) -> str:
        """The algorithm that was used to created the signature."""
        return self._jose.alg

    @property
    def cty(self) -> str:
        """The content type included in the JWT."""
        return self._jose.cty

    @property
    def kid(self) -> str:
        """Identifies the key that was used to create the signature."""
        return self._jose.kid

    @property
    def typ(self) -> str:
        """The type of object as specified by the ``typ`` claim in the
        header.
        """
        return self._jose.typ

    @property
    def protected(self) -> JOSEHeader:
        """The protected JWS header."""
        return self._protected

    @classmethod
    async def create(
            cls,
            keychain: Keychain,
            algorithm: str,
            using: str,
            payload: bytes,
            protected: bytes = None,
            header: dict = None
        ) -> 'Signature':
        """Create and sign a :class:`Signature` for a payload.

        Be aware that the `algorithm` parameter must match the ``alg``
        claim specified on the JWS Protected Header.

        The signature is calculated as follows (from :rfc:`7515`, section 5.1):

            Compute the encoded header value BASE64URL(UTF8(JWS Protected
            Header)).  If the JWS Protected Header is not present (which can
            only happen when using the JWS JSON Serialization and no
            "protected" member is present), let this value be the empty
            string .... Compute the JWS Signature in the manner defined for the
            particular algorithm being used over the JWS Signing Input
            ASCII(BASE64URL(UTF8(JWS Protected Header)) || '.' ||
            BASE64URL(JWS Payload)).  The "alg" (algorithm) Header Parameter
            MUST be present in the JOSE Header, with the algorithm value
            accurately representing the algorithm used to construct the JWS
            Signature.

        The resulting :class:`Signature` may then be added to an existing
        or a new :class:`JSONWebSigature` instance.
        """
        if algorithm != "none":
            signature = await keychain.sign(
                data=bytes.join(b'.', [protected or b'', payload]),
                using=using,
                algorithm=algorithm
            )
        else:
            signature = b''
        return cls(
            signature=signature,
            protected=protected,
            header=header
        )

    def __init__(
        self,
        signature: bytes,
        protected: bytes = None,
        header: dict = None
    ):
        """Per :rfc:`7515`:

            protected

            The "protected" member MUST be present and contain the value
            BASE64URL(UTF8(JWS Protected Header)) when the JWS Protected
            Header value is non-empty; otherwise, it MUST be absent.  These
            Header Parameter values are integrity protected.
            
            header

            The "header" member MUST be present and contain the value JWS
            Unprotected Header when the JWS Unprotected Header value is non-
            empty; otherwise, it MUST be absent.  This value is represented as
            an unencoded JSON object, rather than as a string.  These Header
            Parameter values are not integrity protected.

            signature

            The "signature" member MUST be present and contain the value
            BASE64URL(JWS Signature).
        """
        self._jose = JOSEHeader.fromb64(protected)
        self._protected = protected
        self._signature = signature
        self._header = header

        # Prevent bypassing signature validation by specifying
        # the `none` algoritm.
        if self.algorithm == "none" and self._signature:
            raise MalformedToken

    def serialize(self) -> bytes:
        """Serialize the signature to URL-encoded base64."""
        return b64encode(self._signature)

    def verify(
        self,
        keychain: Keychain,
        payload: bytes,
        using: str = None,
        allow_none: bool = False
    ) -> bool:
        """Verify that the :class:`Signature` is valid for the given `payload`."""
        if not allow_none and self.algorithm == "none":
            raise AlgorithmNotAllowed
        return keychain.verify(
            signature=self._signature,
            data=self._protected + b'.' + payload,
            using=using or self.kid,
            algorithm=self.algorithm
        )

    def __bytes__(self) -> bytes:
        return self._signature
