"""Declares :class:`TrustedIssuers`."""
import asyncio
import typing

from ckms import Keychain
from .claimset import ClaimSet
from .jsonwebsignature import JSONWebSignature
from .payloadcodec import PayloadCodec
from .exceptions import JOSEException
from .exceptions import InvalidSignature # type: ignore
from .models import JSONWebKey
from .models import JSONWebKeySet


class TrustedIssuers:
    """Maintains a registry of trusted issuers and their signing/encryption
    keys.
    """
    deferred: object = object()
    issuers: dict[str, JSONWebKeySet]
    trusted: set[str]
    InvalidSignature: type[JOSEException] = InvalidSignature
    UntrustedIssuer: type[Exception] = type('UntrustedIssuer', (Exception,), {})

    def __init__(self, trusted: set[str] = None):
        self.issuers = {}
        self.trusted = trusted or set()

    def get_issuer_jwks(
        self,
        issuer: str,
    ) -> JSONWebKeySet:
        """Return the JWKS of the issuer."""
        return self.issuers[issuer]

    def is_trusted(self, issuer: str | None) -> bool:
        """Return a boolean indicating if the given issuer is trusted."""
        return issuer in self.issuers

    def trust(
        self,
        issuer: str,
        keys: list[JSONWebKey] | JSONWebKeySet | Keychain | None = None
    ) -> None:
        """Add a trusted issuer to the registry.

        If the `keys` parameter is provided, no attempt is made to fetch
        the issuers' keys using any mechanism.
        """
        self.trusted.add(issuer)
        if keys is not None:
            if isinstance(keys, list):
                keys = JSONWebKeySet(keys=keys)
            self.issuers[issuer] = typing.cast(JSONWebKeySet, keys)

    async def setup(self) -> None:
        """Fetch the remote keys for all issuers and setup the
        :class:`~ckms.jose.models.JSONWebKeySet` instances.
        """
        await asyncio.gather(*[self.setup_issuer(x) for x in self.trusted])

    async def setup_issuer(self, issuer: str) -> None:
        """Setup the previously registered issuer `issuer`."""
        jwks = self.issuers.get(issuer)
        if isinstance(jwks, Keychain):
            await jwks.setup()
            jwks = JSONWebKeySet.fromkeychain(jwks)
        elif jwks is None:
            jwks = await JSONWebKeySet.discover(issuer)
        self.issuers[issuer] = jwks

    async def verify_token(
        self,
        token: bytes | str,
        decrypter: Keychain | None = None,
        accept: set[str] = {"at+jwt", "jwt"}
    ) -> ClaimSet:
        """Verify a token that was issued by one of the trusted issuers. The token
        is either a JWT or Nested JWT and must specify the `iss` claim.
        """
        if isinstance(token, str):
            token = str.encode(token , 'ascii')
        codec = PayloadCodec(decrypter=decrypter)
        jws = typing.cast(
            JSONWebSignature,
            await codec.decode(
                token,
                _type=JSONWebSignature
            )
        )
        if str.lower(jws.typ or '') not in accept:
            raise TypeError("Unacceptable payload.")
        return await self.verify_jws(jws)

    async def verify_jws(self, jws: JSONWebSignature) -> ClaimSet:
        """Verifies a :class:`ckms.jose.JSONWebSignature` instance and
        return the claims it contains. The claims are not validated."""
        claims = jws.get_claimset()
        if not self.is_trusted(claims.iss):
            raise self.UntrustedIssuer
        assert claims.iss is not None # nosec
        jwks = self.get_issuer_jwks(claims.iss)
        if not await jws.verify(jwks): # type: ignore
            raise self.InvalidSignature
        return claims
