"""Declares :class:`EncryptOperation`."""
import typing

from cryptography.hazmat.primitives import hashes

from .operation import Operation
from .plaintext import Plaintext


class EncryptOperation(Operation):
    __module__: str = 'ckms.ops'
    supported_input: tuple = (Plaintext,)

    @property
    def algorithm(self) -> hashes.Hash:
        """Return a :class:`cryptography.hazmat.primitives.hashes.Hash`
        implementation representing the hashing algorithm used
        for this :class:`EncryptOperation`.
        """
        return self._algorithm # pragma: no cover

    @property
    def padding(self) -> str:
        """Identifies the padding scheme used with the signing
        operation, or ``None`` if no padding is used.
        """
        return self._padding

    @property
    def pt(self) -> Plaintext:
        """Return a :class:`Plaintext` instance."""
        return self._data

    def __init__(
        self,
        data: Plaintext,
        iv: int = 12,
        tag: int = 128,
        mode: str = None,
        padding: str = None,
        algorithm: str = None,
        key_size: int = None
    ):
        super().__init__(data)
        self.iv = iv
        self.mode = mode
        self.aad = tag
        self.key_size = key_size
        self._padding = padding
        self._algorithm = algorithm

    async def execute(
        self,
        provider: 'Provider',
        key: 'KeySpecification'
    ) -> typing.Any:
        """Execute the operation using the given `provider` and `key`."""
        return await provider.encrypt(key, self)
