"""Declares :class:`Message`."""
import hashlib
import typing

from ckms.utils import AwaitableWrapper
from .signable import Signable


class Message(Signable):
    """A :class:`Signable` implementation that represents input
    data that is kept in memory.
    """
    __module__: str = 'unimatrix.ext.kms.operations'

    async def get_data(self) -> bytes:
        """Return a byte-sequence holding the input data."""
        return self.data

    def get_digest(
        self,
        algorithm: str,
        awaitable: bool = True
    ) -> typing.Union[bytes, AwaitableWrapper]:
        """Return a digest of the input data using the hashing algorithm
        specified by the `algorithm` parameter.
        """
        if awaitable:
            digest = AwaitableWrapper(self._get_digest, algorithm)
        else:
            digest = self._get_digest(algorithm)
        return digest

    def _get_digest(
        self,
        algorithm: str
    ) -> bytes:
        if algorithm not in {'sha256', 'sha384', 'sha512'}:
            raise ValueError(f"Invalid hashing algorithm: {algorithm}")
        hasher = getattr(hashlib, algorithm)(self.data)
        return hasher.digest()

    def __bytes__(self) -> bytes:
        return self.data
