"""Declares :class:`Operation`."""
import typing

from ckms.utils import AwaitableWrapper
from httpcore import stream

from .data import Data
from .message import Message


class Operation:
    """The base class for all cryptographic operations."""
    __module__: str = 'ckms.ops'
    supported_input = tuple()

    @property
    def data(self) -> typing.Union[bytes, Data]:
        return self._data

    @classmethod
    def frombytes(cls, data: bytes, **kwargs):
        """Create a new operation from a message."""
        return cls(Message(data), **kwargs)

    def __init__(self, data: Data):
        assert isinstance(data, self.supported_input), repr(data) # nosec
        self._data = data

    async def execute(
        self,
        provider: 'Provider',
        key: 'KeySpecification'
    ) -> typing.Any:
        """Execute the operation using the given `provider` and `key`."""
        raise NotImplementedError

    def _get_digest(
        self,
        algorithm: str,
        awaitable: bool = True
    ) -> typing.Union[bytes, AwaitableWrapper]:
        return self._data.get_digest(algorithm, awaitable=awaitable)

    async def get_data(self) -> bytes:
        """Return the input data for the :class:`Operation` i.e. the
        message or the plain text.
        """
        return await self._data.get_data()

    def process_result(self, result: bytes) -> bytes:
        """Hook to modify the result of an operation."""
        return result
