"""Declares :class:`SignOperation`."""
import typing

from cryptography.hazmat.primitives import hashes

from .digest import Digest
from .message import Message
from .operation import Operation



class SignOperation(Operation):
    __module__: str = 'ckms.ops'
    supported_input: tuple = (
        Digest,
        Message
    )

    @property
    def algorithm(self) -> hashes.Hash:
        """Return a :class:`cryptography.hazmat.primitives.hashes.Hash`
        implementation representing the hashing algorithm used
        for this :class:`SignOperation`.
        """
        return self._algorithm

    @property
    def padding(self) -> str:
        """Identifies the padding scheme used with the signing
        operation, or ``None`` if no padding is used.
        """
        return self._padding

    def __init__(
        self,
        data: typing.Union[Digest, Message],
        padding: typing.Optional[str] = None,
        algorithm: typing.Optional[hashes.Hash] = None,
        prehashed: bool = True
    ):
        super().__init__(data)
        self._data = data
        self._padding = padding
        self._algorithm = algorithm
        self._prehashed = prehashed

    def get_hashing_algorithm(self) -> str:
        """Return a string containing the name of the hashing algorithm."""
        return self.algorithm.name # type: ignore

    async def execute(
        self,
        provider: 'Provider',
        key: 'KeySpecification'
    ) -> typing.Any:
        """Execute the operation using the given `provider` and `key`."""
        return await provider.sign(key, self)

    def get_digest(self, awaitable: bool = True) -> bytes:
        """Return the digest to calculate a signature over."""
        return self._get_digest(
            self.algorithm.name,
            awaitable=awaitable
        )