""".. _guides-implementing-a-provider:

=======================
Implementing a provider
=======================
"""
import importlib
import typing

from .cryptography import CryptographyProvider
from .exceptions import ProviderDoesNotExist
from .keyspecification import KeySpecification
from .provider import Provider
from .random import RandomProvider


__all__ = [
    'register',
    'CryptographyProvider',
    'KeySpecification',
    'Provider',
    'ProviderDoesNotExist',
    'RandomProvider'
]


# Maintains a mapping of symbolic names to the qualified names
# of a provider implementation.
_providers: dict[str, tuple[str, dict[str, typing.Any]]] = {
    'aws': ("ckms.providers.aws:AmazonProvider", {}),
    'azure': ("ckms.providers.azure:AzureProvider", {}),
    'local': ("ckms.providers.local:LocalProvider", {}),
    'google': ("ckms.providers.google:GoogleProvider", {}),
    'yubikey': ("ckms.providers.yubikey:YubikeyProvider", {}),
    'yubihsm': ("ckms.providers.yubihsm:YubiHSMProvider", {}),
}


def register(
    name: str,
    qualname: str,
    **kwargs: typing.Any
) -> None:
    """Registers a new provider under an alias. This function
    may be used to alias a custom provider implementation for
    a backend that is not supported by :mod:`ckms`, or to
    use a different backend implementation.
    """
    _providers[name] = (qualname, kwargs)


def get(alias: str, **kwargs: typing.Any) -> None:
    """Get an instance of the given provider."""
    if alias not in _providers:
        raise ProviderDoesNotExist(alias)
    qualname, defaults = _providers[alias]
    defaults.update(kwargs)
    module_name, class_name = str.split(qualname, ':')
    module = importlib.import_module(module_name)
    return getattr(module, class_name)(**defaults)


def get_available() -> typing.List[str]:
    """Return the list of available providers."""
    return list(sorted(_providers.keys()))