"""Declares :class:`CryptographyProvider`."""
import hmac
import os
import typing

import cryptography.exceptions
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import ed448
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import utils
from cryptography.hazmat.primitives import ciphers
from cryptography.hazmat.primitives.keywrap import aes_key_wrap
from cryptography.hazmat.primitives.keywrap import aes_key_unwrap

from ckms import ops
from ckms.utils import AwaitableWrapper
from ckms.utils import bytes_to_number
from ckms.utils import number_to_bytes


class CryptographyProvider:
    """Mixin class to expose cryptographic operations implemented
    by the :mod:`cryptography` package.
    """
    __module__: str = 'ckms.providers'
    InvalidSignature = cryptography.exceptions.InvalidSignature

    async def decrypt_aes(
        self,
        key: bytes,
        op: ops.DecryptOperation
    ) -> bytes:
        ct = await op.get_data()
        data = bytes(ct)
        if op.mode == 'GCM':
            assert ct.iv is not None
            assert ct.tag is not None
            assert ct.aad is not None
            c = ciphers.Cipher(
                ciphers.algorithms.AES(bytes(key)),
                ciphers.modes.GCM(ct.iv, ct.tag)
            )
            d = c.decryptor()
            d.authenticate_additional_data(ct.aad)
        elif op.mode == "CBC":
            c = ciphers.Cipher(
                ciphers.algorithms.AES(bytes(key)),
                ciphers.modes.CBC(ct.iv)
            )
            d = c.decryptor()
        else:
            raise NotImplementedError(op.mode)
        pt = d.update(data) + d.finalize()
        if op.mode == "CBC":
            p = op.padding.unpadder()
            pt = p.update(pt) + p.finalize()
        return pt

    async def decrypt_rsa(
        self,
        key: typing.Union[rsa.RSAPrivateKey],
        op: ops.EncryptOperation
    ) -> bytes:
        return key.decrypt(
            bytes(await op.get_data()),
            padding=op.padding
        )

    async def encrypt_aes(
        self,
        key: bytes,
        op: ops.EncryptOperation
    ) -> ops.Ciphertext:
        pt = op.pt
        iv = os.urandom(op.iv)
        aad = pt.aad if pt.aad is not None else os.urandom(op.aad)
        if op.mode == 'GCM':
            c = ciphers.Cipher(
                ciphers.algorithms.AES(bytes(key)),
                ciphers.modes.GCM(iv)
            )
            e = c.encryptor()
            if aad:
                e.authenticate_additional_data(aad)
            data = bytes(pt)
        elif op.mode == "CBC":
            c = ciphers.Cipher(
                ciphers.algorithms.AES(bytes(key)),
                ciphers.modes.CBC(iv)
            )
            p = op.padding.padder()
            e = c.encryptor()
            data = p.update(bytes(pt)) + p.finalize()
        else:
            raise NotImplementedError(op.mode)
        ct = e.update(data) + e.finalize()
        return ops.Ciphertext(
            data=ct,
            aad=aad,
            iv=iv,
            tag=getattr(e, 'tag', None)
        )

    async def encrypt_rsa(
        self,
        key: typing.Union[rsa.RSAPrivateKey, rsa.RSAPublicKey],
        op: ops.EncryptOperation
    ) -> bytes:
        if isinstance(key, rsa.RSAPrivateKey):
            key = key.public_key()
        return key.encrypt(
            await op.get_data(),
            padding=op.padding
        )

    async def sign_ec(
        self,
        key: ec.EllipticCurvePrivateKey,
        op: ops.SignOperation
    ) -> bytes:
        n = (key.curve.key_size + 7) // 8
        sig =key.sign(
            data=await op.get_digest(),
            signature_algorithm=ec.ECDSA(utils.Prehashed(op.algorithm))
        )
        r, s = utils.decode_dss_signature(sig)
        return number_to_bytes(r, n) + number_to_bytes(s, n)

    async def sign_eddsa(
        self,
        key: typing.Union[ed448.Ed448PrivateKey, ed25519.Ed25519PrivateKey],
        op: ops.SignOperation
    ) -> bytes:
        return key.sign(await op.get_data())

    async def sign_hmac(
        self,
        key: bytes,
        op: ops.SignOperation
    ) -> bytes:
        return hmac.digest(key, bytes(op.data), op.get_hashing_algorithm())

    async def sign_rsa(
        self,
        key: rsa.RSAPrivateKey,
        op: ops.SignOperation
    ) -> bytes:
        """Handle a :class:`SignOperation` using an RSA key."""
        return key.sign(
            data=await op.get_digest(),
            padding=op.padding,
            algorithm=utils.Prehashed(op.algorithm)
        )

    def verify_ec(
        self,
        key: typing.Union[ec.EllipticCurvePrivateKey, ec.EllipticCurvePublicKey],
        op: ops.VerifyOperation
    ) -> bool:
        if not isinstance(key, ec.EllipticCurvePublicKey):
            key = key.public_key()
        n = (key.curve.key_size + 7) // 8
        key.verify(
            signature=utils.encode_dss_signature(
                bytes_to_number(op.signature[:n]),
                bytes_to_number(op.signature[n:]),
            ),
            data=op.get_digest(awaitable=False),
            signature_algorithm=ec.ECDSA(utils.Prehashed(op.algorithm))
        )
        return True

    def verify_eddsa(
        self,
        key: typing.Union[
            ed448.Ed448PrivateKey,
            ed448.Ed448PublicKey,
            ed25519.Ed25519PrivateKey,
            ed25519.Ed25519PublicKey
        ],
        op: ops.VerifyOperation
    ) -> bool:
        if not isinstance(key, (ed448.Ed448PublicKey, ed25519.Ed25519PublicKey)):
            key = key.public_key()
        key.verify(
            signature=op.signature,
            data=bytes(op.data),
        )
        return True

    def verify_hmac(
        self,
        key: bytes,
        op: ops.VerifyOperation
    ) -> bool:
        return hmac.compare_digest(
            op.signature,
            hmac.digest(key, bytes(op.data), op.get_hashing_algorithm())
        )

    def verify_rsa(
        self,
        key: typing.Union[rsa.RSAPrivateKey, rsa.RSAPublicKey],
        op: ops.VerifyOperation
    ) -> bool:
        if not isinstance(key, rsa.RSAPublicKey):
            key = key.public_key()
        key.verify(
            signature=op.signature,
            data=op.get_digest(awaitable=False),
            padding=op.padding,
            algorithm=utils.Prehashed(op.algorithm)
        )
        return True

    def wrap_key(
        self,
        spec: 'KeySpecification',
        op: ops.KeyWrapOperation,
        awaitable: bool = True
    ) -> typing.Union[bytes, AwaitableWrapper]:
        return AwaitableWrapper.call(
            awaitable=awaitable,
            func=aes_key_wrap,
            wrapping_key=bytes(spec.key),
            key_to_wrap=bytes(op.data)
        )

    def unwrap_key(
        self,
        spec: 'KeySpecification',
        op: ops.KeyUnwrapOperation,
        awaitable: bool = True
    ) -> typing.Union[bytes, AwaitableWrapper]:
        return AwaitableWrapper.call(
            awaitable=awaitable,
            func=aes_key_unwrap,
            wrapping_key=bytes(spec.key),
            wrapped_key=bytes(op.data)
        )

    def _verify(
        self,
        spec: 'KeySpecification',
         op: ops.VerifyOperation
    ) -> bool:
        try:
            if spec.kty == 'RSA':
                result = self.verify_rsa(spec.key, op)
            elif spec.kty == "EC":
                result = self.verify_ec(spec.key, op)
            elif spec.kty == "OKP":
                result = self.verify_eddsa(spec.key, op)
            elif spec.kty == "oct":
                # assume hmac.
                k = spec.key
                result = self.verify_hmac(bytes(k), op)
            else:
                raise NotImplementedError(spec.kty, spec.key)
        except self.InvalidSignature:
            result = False
        return result