"""Declares :class:`GoogleProvider`."""
import asyncio
import atexit

import crcmod
import google.auth
from google.cloud.kms import AsymmetricSignRequest
from google.cloud.kms import KeyManagementServiceAsyncClient

import ckms.utils
from ckms import ops
from ckms import providers

from .keyspecification import KeySpecification


class GoogleProvider(providers.Provider):
    __module__: str = 'ckms.providers.google'
    spec_model: type = KeySpecification
    lock: asyncio.Lock = asyncio.Lock()

    @staticmethod
    def crc32c(data: bytes) -> int:
        """Calculates the CRC32C checksum of the provided data."""
        func = crcmod.predefined.mkPredefinedCrcFun('crc-32c')
        return func(data)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.credentials, _ = google.auth.default()

    def get_kms_client(self):
        """Return a :class:`google.cloud.kms.KeyManagementServiceAsyncClient`
        instance required with the credentials established by the
        :class:`GoogleProvider` object.
        """
        return KeyManagementServiceAsyncClient(credentials=self.credentials)

    async def sign(self, spec: 'KeySpecification', op: ops.SignOperation) -> bytes:
        """Perform a signing operation with the parameters specified
        by :class:`~ckms.ops.SignOperation` `op`.
        """
        async with self.get_kms_client() as client:
            if spec.is_asymmetric():
                data = await op.get_digest()
                checksum = self.crc32c(data)
                response = await client.asymmetric_sign(
                    request={
                        'name': spec.name,
                        'digest': {
                            op.algorithm.name: data
                        },
                        'digest_crc32c': checksum
                    },
                    timeout=10
                )
                sig = response.signature
                crc = response.signature_crc32c
                if not response.verified_digest_crc32c:
                    raise ValueError("Digest checksum not considered or verified by cloud service.")
            else:
                # Here is assumed that the algorithm is HMAC_SHA256.
                data = await op.get_data()
                checksum = self.crc32c(data)
                response = await client.mac_sign(
                    request={
                        'name': spec.name,
                        'data': data,
                        'data_crc32c': checksum
                    },
                    timeout=10
                )
                sig = response.mac
                crc = response.mac_crc32c
                if not response.verified_data_crc32c:
                    raise ValueError("Data checksum not considered or verified by cloud service.")

        if crc != self.crc32c(sig):
            raise ValueError("Corrupted signature in transit.")

        if spec.kty == "EC":
            sig = ckms.utils.normalize_ec_signature(
                l=(spec.key_size + 7) // 8,
                sig=sig
            )
        return sig

    async def verify(
        self,
        spec: 'KeySpecification',
        op: ops.VerifyOperation
    ) -> bool:
        """Handle a :class:`VerifyOperation` using the given key configuration."""
        if spec.is_asymmetric():
            result = await self.keychain.verify(
                signature=op.signature,
                data=op.data,
                using=spec.public_keys[0],
                algorithm=spec.alg
            )
        else:
            async with self.get_kms_client() as client:
                data = await op.get_data()
                checksum = self.crc32c(data)
                response = await client.mac_verify(
                    request={
                        'name': spec.name,
                        'data': data,
                        'data_crc32c': checksum,
                        'mac': op.signature,
                        'mac_crc32c': self.crc32c(op.signature)
                    },
                    timeout=10
                )
                if not response.verified_data_crc32c:
                    raise ValueError("Data checksum not considered or verified by cloud service.")
                if not response.verified_mac_crc32c:
                    raise ValueError("Mac checksum not considered or verified by cloud service.")
                if response.verified_success_integrity != response.success:
                    raise ValueError("Result integrity verification failed.")
                result = response.success

        return result