"""Declares :class:`KeySpecification`."""
import asyncio
import typing

from cryptography.hazmat.primitives.serialization import load_pem_public_key
from google.cloud.kms import CryptoKey
from google.cloud.kms import CryptoKeyVersion
from google.cloud.kms import KeyManagementServiceAsyncClient

from ckms import providers
from ckms import Keychain


ALGORITHM_MAP = {
    'EC_SIGN_P256_SHA256'           : ('EC', 'ES256', 'P-256', 'sig', 256),
    'EC_SIGN_P384_SHA384'           : ('EC', 'ES384', 'P-384', 'sig', 384),
    'EC_SIGN_SECP256K1_SHA256'      : ('EC', 'ES256K', 'P-256K', 'sig', 256),
    'RSA_SIGN_PSS_2048_SHA256'      : ('RSA', 'PS256', None, 'sig', 2048),
    'RSA_SIGN_PSS_3072_SHA256'      : ('RSA', 'PS256', None, 'sig', 3072),
    'RSA_SIGN_PSS_4096_SHA256'      : ('RSA', 'PS256', None, 'sig', 4096),
    'RSA_SIGN_PSS_4096_SHA512'      : ('RSA', 'PS512', None, 'sig', 4096),
    'RSA_SIGN_PKCS1_2048_SHA256'    : ('RSA', 'RS256', None, 'sig', 2048),
    'RSA_SIGN_PKCS1_3072_SHA256'    : ('RSA', 'RS256', None, 'sig', 3072),
    'RSA_SIGN_PKCS1_4096_SHA256'    : ('RSA', 'RS256', None, 'sig', 4096),
    'RSA_SIGN_PKCS1_4096_SHA512'    : ('RSA', 'RS512', None, 'sig', 4096),
    'HMAC_SHA256'                   : ('oct', 'HS256', None, 'sig', None),
    'RSA_DECRYPT_OAEP_2048_SHA1'    : ('RSA', 'RSA-OAEP', None, 'enc', 2048),
    'RSA_DECRYPT_OAEP_2048_SHA256'  : ('RSA', 'RSA-OAEP-256', None, 'enc', 2048),
    'RSA_DECRYPT_OAEP_3072_SHA1'    : ('RSA', 'RSA-OAEP', None, 'enc', 3072),
    'RSA_DECRYPT_OAEP_3072_SHA256'  : ('RSA', 'RSA-OAEP-256', None, 'enc', 3072),
    'RSA_DECRYPT_OAEP_4096_SHA1'    : ('RSA', 'RSA-OAEP', None, 'enc', 4096),
    'RSA_DECRYPT_OAEP_4096_SHA256'  : ('RSA', 'RSA-OAEP-256', None, 'enc', 4096),
    'RSA_DECRYPT_OAEP_4096_SHA512'  : ('RSA', 'RSA-OAEP-512', None, 'enc', 4096),
    'GOOGLE_SYMMETRIC_ENCRYPTION'   : ('oct', 'A256GCM', None, 'enc', 256)
}


class KeySpecification(providers.KeySpecification):
    """A :class:`ckms.providers.KeySpecification` implementation that resolves
    to key material that is maintained by Google Cloud KMS.
    """
    __module__: str = 'ckms.providers.google'
    project: str
    location: str
    keyring: str
    key: str

    # State
    crypto_key: typing.Optional[CryptoKey] = None
    versions: list = []
    public_keys: list = []
    key_size: typing.Optional[int] = None

    class Config:
        arbitrary_types_allowed: bool = True

    @property
    def name(self):
        return self.versions[0].name

    def is_asymmetric(self) -> bool:
        """Return a boolean indicating if the key is an asymmetric
        algorithm.
        """
        return self.crypto_key.purpose.name in {'ASYMMETRIC_DECRYPT', 'ASYMMETRIC_SIGN'}

    async def load(self) -> None:
        """Loads key material and metadata from the cryptographic backend."""
        async with self.provider.get_kms_client() as client:
            await self.load_versions(client, self.keychain)
        self.loaded = True
        return self

    async def load_versions(
        self,
        client: KeyManagementServiceAsyncClient,
        keychain: Keychain
    ) -> str:
        """Return a string containing the qualified name to the crypto
        key version at the Google Cloud KMS keyring.
        """
        path = KeyManagementServiceAsyncClient.crypto_key_path(
            project=self.project,
            location=self.location,
            key_ring=self.keyring,
            crypto_key=self.key
        )
        self.crypto_key = key = await client.get_crypto_key(
            request={
                'name': path
            },
            timeout=10
        )

        response = client.list_crypto_key_versions(
            request={
                'parent': path,
                'filter': "state=ENABLED"
            },
            timeout=10
        )

        futures = []
        async for ref in await response:
            request = {'name': ref.name}
            futures.append(client.get_crypto_key_version(request=request, timeout=10))
        self.versions.extend(await asyncio.gather(*futures))

        # Sort version by timestamp so that the most recent version
        # is the first member.
        self.versions = list(
            sorted(self.versions, key=lambda x: -x.generate_time.timestamp())
        )

        # Get the algorithm, use and curve (if an elliptic curve key)
        # and set as attributes.
        self.kty, self.alg, self.crv, self.use, self.key_size = ALGORITHM_MAP[self.versions[0].algorithm.name]

        # Get public keys for asymmetric algorithms and load them using the
        # local provider.
        if self.is_asymmetric():
            local = keychain.get_provider('local')
            for version in self.versions:
                public = await self.load_version(client, keychain, local, version)
                if not self.public_kid:
                    self.public_kid = self.calculate_kid(public.key)
                keychain.alias(public.kid, self.kid) # TODO
                self.public_keys.append(public.kid)

    async def load_version(
        self,
        client: KeyManagementServiceAsyncClient,
        keychain: Keychain,
        provider: providers.Provider,
        version: CryptoKeyVersion
    ) -> 'KeySpecification':
        """Load a key version and the corresponding public key."""
        request = {'name': version.name}
        response = await client.get_public_key(request=request, timeout=10)
        public = await provider.load(
            kid=None,
            spec={
                'key': load_pem_public_key(str.encode(response.pem, 'ascii')),
                'keychain': self.keychain
            }
        )
        public.kty, public.alg, public.crv, public.use, _ = ALGORITHM_MAP[version.algorithm.name]
        keychain.trust(public)
        return public
