"""Declares :class:`KeySpecification`."""
import hashlib
import typing

import pydantic
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import ed448
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization

from ckms.core import KeyInspector


CURVE_CLASSES = {
    'P-256': ec.SECP256R1,
    'P-256K': ec.SECP256K1,
    'P-384': ec.SECP384R1,
    'P-521': ec.SECP521R1
}

PUBLIC_KEY_TYPES = (
    ec.EllipticCurvePublicKey,
    ed448.Ed448PublicKey,
    ed25519.Ed25519PublicKey,
    rsa.RSAPublicKey
)

PRIVATE_KEY_TYPES = (
    ec.EllipticCurvePrivateKey,
    ed448.Ed448PrivateKey,
    ed25519.Ed25519PrivateKey,
    rsa.RSAPrivateKey
)


class KeySpecification(pydantic.BaseModel):
    """The base class and interface for all key specification implementations."""
    __module__: str = 'ckms.provider'

    keychain: typing.Any
    provider: typing.Any
    loaded: typing.Optional[bool] = False
    bit_length: typing.Optional[int] = None
    public_keys: typing.List[str] = []
    trust: bool = False
    public_kid: typing.Optional[str] = None
    private_kid: typing.Optional[str] = None
    tags: typing.Optional[typing.List[str]] = []
    weight: typing.Optional[int] = 0

    # JWK fields
    kid: typing.Optional[str] = None
    kty: typing.Optional[str] = None
    use: typing.Optional[str] = None
    alg: typing.Optional[str] = None
    crv: typing.Optional[str] = None

    @staticmethod
    def calculate_kid(
        key: typing.Union[PRIVATE_KEY_TYPES+PUBLIC_KEY_TYPES]
    ) -> str:
        """Calculate a key identifier based on the hash of the public numbers
        or public key or an asymmetric key.
        """
        return KeyInspector().calculate_kid(key)

    def get_curve(self) -> ec.EllipticCurve:
        """Return the :class:`cryptography.hazmat.primitives.asymmetric.ec.EllipticCurve`
        specified by the :attr:`crv` attribute.
        """
        return CURVE_CLASSES[self.crv]()

    def get_private_kid(self) -> str:
        """Return the private key identifier of this key."""
        return self.private_kid

    def is_loaded(self) -> bool:
        """Return a boolean indicating if the key is loaded."""
        return self.loaded

    async def load(self) -> None:
        """Loads key material and metadata from the cryptographic backend."""
        raise NotImplementedError

    async def _load(self) -> 'KeySpecification':
        await self.load()
        return self

    @staticmethod
    def _get_kty(key) -> str:
        if isinstance(key, (
            ed448.Ed448PrivateKey,
            ed448.Ed448PublicKey,
            ed25519.Ed25519PrivateKey,
            ed25519.Ed25519PublicKey
        )):
            return 'OKP'
        elif isinstance(key, (ec.EllipticCurvePrivateKey, ec.EllipticCurvePublicKey)):
            return 'EC'
        elif isinstance(key, (rsa.RSAPrivateKey, rsa.RSAPublicKey)):
            return 'RSA'
        else:
            return 'oct'

    def __await__(self):
        return self._load().__await__()