# pylint: skip-file
import secrets
import typing

from ckms import ops
from ckms import providers
from ckms.utils import AwaitableWrapper

from .cipher import Cipher
from .hmacsecret import HMACSecret
from .keyspecification import KeySpecification


class LocalProvider(providers.Provider, providers.CryptographyProvider):
    """A :class:`ckms.providers.Provider` implementations that uses
    key material from the local system.
    """
    __module__: str = 'ckms.providers.local'
    spec_model: type = KeySpecification
    awaitable: bool = True

    def decrypt(
        self,
        spec: 'KeySpecification',
        op: ops.DecryptOperation
    ) -> typing.Union[bytes, typing.Awaitable[bytes]]:
        """Perform a decryption operation with the parameters specified
        by :class:`~ckms.ops.DecryptOperation` `op`.
        """
        if spec.kty == 'RSA':
            func = self.decrypt_rsa
        elif spec.kty == "oct":
            func = self.decrypt_aes
        else:
            raise NotImplementedError
        return AwaitableWrapper.call(
            awaitable=self.awaitable,
            func=func,
            key=spec.key,
            op=op
        )

    async def encrypt(self,
        spec: 'KeySpecification',
        op: ops.EncryptOperation
    ) -> typing.Union[bytes, ops.Ciphertext]:
        """Perform an encryption operation with the parameters specified
        by :class:`~ckms.ops.EncryptOperation` `op`.
        """
        if spec.kty == 'RSA':
            result = await self.encrypt_rsa(spec.key, op)
        elif spec.kty in {"EC", "EdDSA"}:
            raise NotImplementedError
        elif spec.kty == "oct":
            result = await self.encrypt_aes(spec.key, op)
        else:
            raise NotImplementedError
        return result

    async def random(self, op: ops.RandomOperation) -> bytes:
        return secrets.token_bytes(op.length)

    async def sign(self, spec: 'KeySpecification', op: ops.SignOperation) -> bytes:
        """Perform a signing operation with the parameters specified
        by :class:`~ckms.ops.SignOperation` `op`.
        """
        if spec.kty == 'RSA':
            result = await self.sign_rsa(spec.key, op)
        elif spec.kty == 'EC':
            result = await self.sign_ec(spec.key, op)
        elif spec.kty == 'OKP':
            result = await self.sign_eddsa(spec.key, op)
        elif isinstance(spec.key, (bytes, HMACSecret)):
            key = spec.key
            if isinstance(key, HMACSecret):
                key = bytes(key)
            result = await self.sign_hmac(key, op)
        else:
            raise NotImplementedError(f"Unsupported key type: {spec.kty}")
        return result

    async def verify(
        self,
        spec: 'KeySpecification',
        op: ops.VerifyOperation
    ) -> bool:
        """Handle a :class:`VerifyOperation` using the given key configuration."""
        return self._verify(spec, op)
