"""Declares :class:`KeySpecification`."""
import typing
import os

import aiofiles
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import ed448
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import x448
from cryptography.hazmat.primitives.asymmetric import x25519
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives.serialization import load_pem_public_key

from ckms import core
from ckms import providers
from .cipher import Cipher
from .hmacsecret import HMACSecret


PUBLIC_KEY_TYPES = (
    ec.EllipticCurvePublicKey,
    ed448.Ed448PublicKey,
    ed25519.Ed25519PublicKey,
    rsa.RSAPublicKey,
    x448.X448PublicKey,
    x25519.X25519PublicKey
)


PRIVATE_KEY_TYPES = (
    ec.EllipticCurvePrivateKey,
    ed448.Ed448PrivateKey,
    ed25519.Ed25519PrivateKey,
    rsa.RSAPrivateKey,
    x448.X448PrivateKey,
    x25519.X25519PrivateKey
)


class KeySpecification(providers.KeySpecification):
    """A :class:`ckms.providers.KeySpecification` implementation that resolves
    to key material that is kept in memory.
    """
    __module__: str = 'ckms.providers.local'
    path: str = None
    secret: typing.Union[bytes, core.RemoteBlob, str] = None
    key: typing.Union[
        ec.EllipticCurvePrivateKey,
        ec.EllipticCurvePublicKey,
        ed448.Ed448PrivateKey,
        ed448.Ed448PublicKey,
        ed25519.Ed25519PrivateKey,
        ed25519.Ed25519PublicKey,
        rsa.RSAPrivateKey,
        rsa.RSAPublicKey,
        x448.X448PrivateKey,
        x448.X448PublicKey,
        x25519.X25519PrivateKey,
        x25519.X25519PublicKey,
        core.RemoteBlob
    ] = None
    length: typing.Optional[int] = None
    use: str

    @staticmethod
    async def _open(path: str, password: str | bytes = None):
        async with aiofiles.open(path, 'rb') as f:
            return self._load(await f.read(), password)

    def _deserialize_key(self, pem: bytes, password: bytes | str | None = None) -> typing.Any:
            if isinstance(password, str):
                password = str.encode(password)
            try:
                return load_pem_private_key(pem, password)
            except ValueError:
                return load_pem_public_key(pem)

    def get_key_material(self) -> typing.Tuple[bytes, bytes]:
        """Extract the raw key material encoded as PEM, in the case of
        asymmetric keys, or the raw octets in the case of HMAC secrets
        or AES keys. Return a tuple containing ``(public, private)``;
        for symmetric algorithms, ``public`` is ``None``.
        """
        if self.kty in {"oct"}:
            return None, bytes(self.key)

        private = public = None
        public_impl = self.key\
            if isinstance(self.key, PUBLIC_KEY_TYPES)\
            else self.key.public_key()
        if not isinstance(self.key, PUBLIC_KEY_TYPES):
            private = self.key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption()
            )
        public = public_impl.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        return public, private

    async def load(self) -> None:
        """Loads key material and metadata from the cryptographic backend."""
        if not (bool(self.path) ^ bool(self.secret) ^ bool(self.key)) and not self.kty: # pragma: no cover
            raise ValueError("Specify either `kty`, `path`, `secret` or `key`.")
        if self.secret: # pragma: no cover
            # Try to determine the key type based on the length of the secret. If the
            # `use` parameter is `enc`, then we assume that it is an AES key (and
            # must have the corresponding length); else a HMAC key is assumed.
            if isinstance(self.secret, core.RemoteBlob):
                self.secret = await core.fetch(self.secret)
            if not isinstance(self.secret, bytes):
                raise NotImplementedError
            elif self.use == 'enc':
                if len(self.secret) not in (16, 24, 32):
                    raise ValueError("Key length must be 16, 24 or 32")
                self.key = Cipher(self.secret)
            else:
                self.key = HMACSecret(self.secret)
        if isinstance(self.key, core.RemoteBlob):
            self.key = self._deserialize_key(await core.fetch(self.key))
        if self.path:
            self.key = await self._open(self.path)
        if self.kty and not self.key:
            # Generate a key. Additional parameters are required.
            if self.kty not in {"EC", "RSA", "OKP", "oct"}:
                raise NotImplementedError
            if self.kty == "RSA" and not self.bit_length: # pragma: no cover
                raise ValueError("The `bit_length` parameter is required when generating keys.")
            if self.kty == "EC" and not self.crv: # pragma: no cover
                raise ValueError("The `crv` parameter is required when generating keys.")

            # Generate the keys.
            if self.kty == "RSA":
                self.key = rsa.generate_private_key(
                    public_exponent=65537,
                    key_size=self.bit_length
                )

            if self.kty == "EC":
                self.key = ec.generate_private_key(self.get_curve())

            if self.kty == "OKP" and self.crv == "Ed448":
                self.key = ed448.Ed448PrivateKey.generate()
                self.alg = "EdDSA"
                self.use = "sig"
            if self.kty == "OKP" and self.crv == "Ed25519":
                self.key = ed25519.Ed25519PrivateKey.generate()
                self.alg = "EdDSA"
                self.use = "sig"
            if self.kty == "OKP" and self.crv == "X448":
                self.key = x448.X448PrivateKey.generate()
                self.use = "enc"
                if not str.startswith(self.alg or '', 'ECDH-ES'):
                    self.alg = "ECDH-ES+A256KW"
            if self.kty == "OKP" and self.crv == "X25519":
                self.key = x25519.X25519PrivateKey.generate()
                self.use = "enc"
                if not str.startswith(self.alg or '', 'ECDH-ES'):
                    self.alg = "ECDH-ES+A256KW"

            if self.kty == "oct":
                self.key = os.urandom(self.length)

        else:
            self.kty = self._get_kty(self.key)

        # Generate a key identifier if the key is public.
        if self.kid is None and isinstance(self.key, PUBLIC_KEY_TYPES):
            self.kid = self.calculate_kid(self.key)

        # If KeySpecification.trust is True, add the public key to the
        # trust registry.
        if isinstance(self.key, PRIVATE_KEY_TYPES):
            public = type(self)(
                keychain=self.keychain,
                provider=self.provider,
                kty=self.kty,
                key=self.key.public_key(),
                trust=True,
                private_kid=self.kid,
                alg=self.alg,
                use=self.use,
                weight=self.weight
            )
            await public.load()
            self.public_keys.append(public)
            assert public.kid == self.calculate_kid(self.key) # nosec
        elif isinstance(self.key, PUBLIC_KEY_TYPES):
            assert self.kid is not None # nosec
            self.keychain.register_public_key(self)

        if isinstance(self.key, PRIVATE_KEY_TYPES):
            self.public_kid = self.calculate_kid(self.key)
            if self.public_kid != self.kid:
                self.keychain.alias(self.kid, self.public_kid)

        self.update_metadata()
        self.loaded = True
        return self

    def update_metadata(self):
        if isinstance(self.key, (ed448.Ed448PrivateKey, ed448.Ed448PublicKey)):
            self.kty = "OKP"
            self.alg = "EdDSA"
            self.crv = "Ed448"
            self.use = "sig"
        elif isinstance(self.key, (ed25519.Ed25519PrivateKey, ed25519.Ed25519PublicKey)):
            self.kty = "OKP"
            self.alg = "EdDSA"
            self.crv = "Ed25519"
            self.use = "sig"
        elif isinstance(self.key, (x448.X448PrivateKey, x448.X448PublicKey)):
            self.kty = "OKP"
            self.crv = "X448"
            self.use = "enc"
            if not str.startswith(self.alg or '', 'ECDH-ES'):
                self.alg = "ECDH-ES+A256KW"
        elif isinstance(self.key, (x25519.X25519PrivateKey, x25519.X25519PublicKey)):
            self.kty = "OKP"
            self.crv = "X25519"
            self.use = "enc"
            if not str.startswith(self.alg or '', 'ECDH-ES'):
                self.alg = "ECDH-ES+A256KW"

    def get_public_keys(self) -> typing.List['KeySpecification']:
        return self.public_keys

    def __hash__(self) -> int:
        return hash(self.kid)

    class Config:
        arbitrary_types_allowed: bool = True
