"""Declares :class:`Provider`."""
import typing

import pydantic

from ckms import ops
from .keyspecification import KeySpecification


class Provider:
    """The base class for all provider implementations. A provider
    may implement one or more of the following cryptographic
    operations:

    - signing
    - verifying
    - encrypting
    - decrypting
    """
    __module__: str = 'ckms.providers'
    spec_model: type = None

    def load(self, kid: str, spec: dict) -> KeySpecification:
        """Load a dictionary representing a key specification into a
        concrete :class:`ckms.provider.KeySpecification` implementation
        used by this provider.
        """
        return self.spec_model(provider=self, kid=kid, **spec)

    async def execute(
        self,
        op: typing.Union[
            ops.SignOperation
        ]
    ) -> typing.Any:
        """The entry point for all cryptographic operations supported
        by the provider.
        """
        if isinstance(op, ops.SignOperation):
            result = await self.sign(op)
        else:
            raise NotImplementedError
        return op.process_result(result)