# pylint: skip-file
"""Utility functions."""
import base64
import binascii
import datetime
import inspect
import typing

from cryptography.hazmat.primitives.asymmetric import utils


def current_timestamp() -> int:
    return int(datetime.datetime.now(datetime.timezone.utc).timestamp())


def b64decode(buf: typing.Union[str, bytes]) -> bytes: # pragma: no cover
    if isinstance(buf, str):
        buf = buf.encode("ascii")
    rem = len(buf) % 4
    if rem > 0:
        buf += b"=" * (4 - rem)
    return base64.urlsafe_b64decode(buf)


def b64decode_int(value: typing.Union[bytes, str]) -> int:
    if isinstance(value, str):
        value = str.encode(value, "ascii")
    return bytes_to_number(b64decode(value))


def b64encode(buf: bytes) -> bytes: # pragma: no cover
    return base64.urlsafe_b64encode(buf).replace(b"=", b"")


def b64encode_int(value: int) -> bytes: # pragma: no cover
    return b64encode(int.to_bytes(value, (value.bit_length() + 7) // 8, 'big'))


def number_to_bytes(value: int, l: int) -> bytes: # pragma: no cover
    padded = str.encode("%0*x" % (2 * l, value), "ascii")
    return binascii.a2b_hex(padded)


def bytes_to_number(value: bytes) -> int: # pragma: no cover
    return int(binascii.b2a_hex(value), 16)


def normalize_ec_signature(l: int, sig: bytes): # pragma: no cover
    r, s = utils.decode_dss_signature(sig)
    return number_to_bytes(r, l) + number_to_bytes(s, l)


class AwaitableWrapper:

    @classmethod
    def call(
        cls,
        awaitable: bool,
        func: typing.Any,
        *args: typing.Any,
        **kwargs: typing.Any
    ):
        return func(*args, **kwargs)\
            if not awaitable else\
            cls(func, *args, **kwargs)

    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    async def get_result(self, result: typing.Any) -> typing.Any:
        return result

    def __await__(self):
        result = self.func(*self.args, **self.kwargs)\
            if callable(self.func)\
            else self.func
        if not inspect.isawaitable(result):
            result = self.get_result(result)
        return result.__await__()
