# pylint: skip-file
import pytest

from ckms import jose
from ckms.jose import JOSEHeader
from ckms.jose.models import JSONWebKey
from .conftest import JOSE_ENCRYPTION_ALGORITHMS
from .conftest import JOSE_SIGNING_ALGORITHMS


class BaseCompatibility:
    """The base class for compatibility tests."""
    skip_algorithms: set = set()

    @pytest.mark.parametrize("using,algorithm", JOSE_SIGNING_ALGORITHMS)
    @pytest.mark.asyncio
    async def test_sign_and_verify_external_compact(self, keychain, codec, using, algorithm):
        """Sign using :mod:`ckms` and verify using the external
        library.
        """
        if algorithm in self.skip_algorithms:
            pytest.skip()
        public, secret = keychain.get_key_material(using)
        jwt = codec.encode(jose.JSONWebToken(foo=1))
        jws = await jwt.sign(
            algorithm=algorithm,
            using=using
        )
        self.verify_compact(public or secret, bytes(jws), algorithm)


    @pytest.mark.parametrize("using,algorithm", JOSE_SIGNING_ALGORITHMS)
    @pytest.mark.asyncio
    async def test_sign_external_and_verify_compact(self, keychain, using, algorithm):
        if algorithm in self.skip_algorithms:
            pytest.skip()
        public, secret = keychain.get_key_material(using)
        assert secret is not None
        token = self.sign_compact(secret, algorithm, using, {
            'foo': 1,
            'bar': 2,
            'baz': 3
        })
        jws = jose.JSONWebSignature.parse(token)
        assert await jws.verify(keychain)

    @pytest.mark.parametrize("using,alg,enc", JOSE_ENCRYPTION_ALGORITHMS)
    @pytest.mark.asyncio
    async def test_jwt_encrypt_external_and_decrypt_compact(self, keychain, codec, using, alg, enc):
        if {alg, enc} & self.skip_algorithms:
            pytest.skip(f"{alg}/{enc} not supported.")
        public, private = keychain.get_key_material(using)
        meta = keychain.get_key_metadata(using)
        spec = keychain.get(using)
        token = self.encrypt_payload_compact(
            JSONWebKey.fromimpl(spec.key, alg=alg),
            public or private,
            alg,
            enc,
            {'kid': meta.get_signing_kid(), 'alg': alg, 'enc': enc, 'typ': "JWT"},
            bytes(jose.JSONWebToken(iss="foo")),
        )
        protected, *_ = bytes.split(token, b'.')
        header = JOSEHeader.fromb64(protected)
        jwt = await codec.decode(token)
        if header.typ == "JWT":
            assert isinstance(jwt, jose.JSONWebToken), repr(jwt)

    @pytest.mark.parametrize("using,alg,enc", JOSE_ENCRYPTION_ALGORITHMS)
    @pytest.mark.asyncio
    async def test_jwt_encrypt_and_decrypt_external_compact(self, keychain, codec, using, alg, enc):
        if {alg, enc} & self.skip_algorithms:
            pytest.skip(f"{alg}/{enc} not supported.")
        public, private = keychain.get_key_material(using)
        spec = keychain.get(using)
        payload = codec.encode(jose.JSONWebToken(iss="foo"))
        token = bytes(await payload.encrypt(algorithm=alg, using=using, encryption=enc))
        header, decrypted = self.decrypt_payload_compact(
            JSONWebKey.fromimpl(spec.key), private, alg, enc, token
        )
        if header is not None:
            assert 'kid' in header
        if decrypted is not None:
            jwt = jose.JSONWebToken.fromb64(decrypted)
            assert jwt.iss == "foo"

    def sign_compact(self, key: bytes, algorithm: str, kid: str, claims: dict) -> bytes:
        raise NotImplementedError

    def encrypt_payload_compact(
        self,
        jwk: 'JSONWebKey',
        key: bytes,
        algorithm: str,
        encryption: str,
        claims: dict,
        payload: bytes
    ) -> bytes:
        pytest.skip("Encryption not implemented or tested")

    def decrypt_payload_compact(
        self,
        jwk: object,
        key: bytes,
        algorithm: str,
        encryption: str,
        payload: bytes
    ) -> tuple:
        pytest.skip("decryption not implemented or tested")

    def verify_compact(self, key: bytes, token: bytes):
        raise NotImplementedError
