# pylint: skip-file
import itertools
import os

import pytest
import pytest_asyncio
from authlib.jose import JsonWebEncryption
from authlib.jose import JsonWebSignature
from cryptography.hazmat.primitives.serialization import load_pem_private_key

import ckms
from ckms import jose
from ckms.utils import b64encode


SIGNING_KEYS = [
    ('rsa', 'RS256'),
    ('rsa', 'RS384'),
    ('rsa', 'RS512'),
    ('rsa', 'PS256'),
    ('rsa', 'PS384'),
    ('rsa', 'PS512'),
]


JOSE_SIGNING_ALGORITHMS = [
    ('rsa', 'RS256'),
    ('rsa', 'RS384'),
    ('rsa', 'RS512'),
    ('rsa', 'PS256'),
    ('rsa', 'PS384'),
    ('rsa', 'PS512'),
    ('hmac', 'HS256'),
    ('hmac', 'HS384'),
    ('hmac', 'HS512'),
    ('p256', 'ES256'),
    ('p384', 'ES384'),
    ('p521', 'ES512'),
    ('p256k', 'ES256K'),
    ('ed448', 'EdDSA'),
    ('ed25519', 'EdDSA'),
]

AES_ENCRYPTION_ALGORITHMS = [
    "A128KW",
    "A192KW",
    "A256KW",
]

RSA_ENCRYPTION_ALGORITHMS = [
    "RSA1_5",
    "RSA-OAEP",
    "RSA-OAEP-256",
    "RSA-OAEP-384",
    "RSA-OAEP-512",
]

ECDH_ENCRYPTION_ALGORITHMS = [
    "ECDH-ES",
    "ECDH-ES+A128KW",
    "ECDH-ES+A192KW",
    "ECDH-ES+A256KW",
]

PBES_ECRYPTION_ALGORITHMS = [
    "PBES2-HS256+A128KW",
    "PBES2-HS384+A192KW",
    "PBES2-HS512+A256KW",
]

JOSE_CEK_ALGORITHMS = [
    #"A128CBC-HS256",
    #"A192CBC-HS384",
    #"A256CBC-HS512",
    "A128GCM",
    "A192GCM",
    "A256GCM",
]


JOSE_ENCRYPTION_ALGORITHMS = (
    list(itertools.product(['rsa'], RSA_ENCRYPTION_ALGORITHMS, JOSE_CEK_ALGORITHMS))
    + list(itertools.product(['aes128'], ["A128KW", "A128GCMKW"], JOSE_CEK_ALGORITHMS))
    + list(itertools.product(['aes192'], ["A192KW"], JOSE_CEK_ALGORITHMS))
    + list(itertools.product(['aes256'], ["A256KW"], JOSE_CEK_ALGORITHMS))
    + list(itertools.product(['aes128'], ['dir'], ["A128GCM"]))
    + list(itertools.product(['aes192'], ['dir'], ["A192GCM"]))
    + list(itertools.product(['aes256'], ['dir'], ["A256GCM"]))
)
    #+ list(itertools.product(['ed448'], ECDH_ENCRYPTION_ALGORITHMS, JOSE_CEK_ALGORITHMS))\
    #+ list(itertools.product(['hmac'], PBES_ECRYPTION_ALGORITHMS, JOSE_CEK_ALGORITHMS))\


@pytest.fixture
def data():
    return os.urandom(16)


@pytest.fixture
def payload():
    return b64encode(b"Hello world!")


@pytest.fixture(scope='session')
def keychain():
    return ckms.Keychain()


@pytest.fixture
def jwt():
    return jose.JSONWebToken(foo=1)


@pytest.fixture
def payload():
    return b'Hello world!'


@pytest_asyncio.fixture
async def encoded_jwt(codec, jwt):
    payload = codec.encode(jwt)
    return bytes(await payload.sign(algorithm='EdDSA', using='ed448'))


@pytest.fixture
def encoded_payload(keychain, payload):
    public, secret = keychain.get_key_material('ed448')
    meta = keychain.get_key_metadata('ed448')
    jws = JsonWebSignature()
    protected = {'alg': 'EdDSA', 'kid': meta.get_signing_kid()}
    return jws.serialize_compact(protected, payload, secret)


@pytest.fixture
def encrypted_jwt(keychain, jwt):
    meta = keychain.get_key_metadata('rsa')
    public, secret = keychain.get_key_material('rsa')
    jwe = JsonWebEncryption()
    protected = {
        'alg': 'RSA-OAEP',
        'enc': 'A256GCM',
        'typ': "JWT",
        'kid': meta.get_signing_kid()
    }
    payload = bytes(jwt)
    return jwe.serialize_compact(protected, payload, public)


@pytest.fixture
def encrypted_payload(keychain, payload):
    meta = keychain.get_key_metadata('rsa')
    public, secret = keychain.get_key_material('rsa')
    jwe = JsonWebEncryption()
    protected = {
        'alg': 'RSA-OAEP',
        'enc': 'A256GCM',
        'kid': meta.get_signing_kid()
    }
    return jwe.serialize_compact(protected, b64encode(payload), public)


@pytest.fixture
def decoder():
    return jose.JOSEDeserializer()


@pytest.fixture
def serializer():
    return jose.JOSEDeserializer()


@pytest_asyncio.fixture(scope='session', autouse=True)
async def setup_local_provider(keychain):
    await ckms.configure({
        'rsa': {
            'provider': "local",
            'path': "pki/rsa.key",
            'trust': True,
            'tags': ['oauth2', 'oauth2.enc'],
            'use': "sig"
        },
        'rsa-fixed': {
            'provider': "local",
            'kty': "RSA",
            'alg': "PS256",
            'path': "pki/rsa-fixed.key",
            'use': "sig"
        },
        'rsa-fixed-enc': {
            'provider': "local",
            'kty': "RSA",
            'alg': "RSA-OAEP-256",
            'path': "pki/rsa-fixed-enc.key",
            'use': "enc"
        },
        'rsa-enc': {
            'provider': "local",
            'path': "pki/rsa-enc.key",
            'tags': ['oauth2', 'oauth2.enc'],
            'use': "enc",
            'weight': -2
        },
        'p256': {
            'provider': "local",
            'path': "pki/p256.key",
            'trust': True,
            'tags': ['oauth2', 'oauth2.sig'],
            'weight': -1,
            'use': 'sig'
        },
        'p384': {
            'provider': "local",
            'path': "pki/p384.key",
            'trust': True
        },
        'p521': {
            'provider': "local",
            'path': "pki/p521.key",
            'trust': True
        },
        'p256k': {
            'provider': "local",
            'path': "pki/p256k.key",
            'trust': True
        },
        'ed448': {
            'provider': "local",
            'path': "pki/ed448.key",
            'trust': True
        },
        'ed25519': {
            'provider': "local",
            'key': load_pem_private_key(open('pki/ed25519.key', 'rb').read(), None),
            'trust': True
        },
        'rsa.public': {
            'provider': "local",
            'path': "pki/rsa.pub"
        },
        'hmac': {
            'provider': "local",
            'secret': os.urandom(64),
            'public_kid': "hmac"
        },
        'aes128': {
            'provider': "local",
            'secret': os.urandom(16),
            'use': "enc",
            'public_kid': 'aes128'
        },
        'aes192': {
            'provider': "local",
            'secret': os.urandom(24),
            'use': "enc",
            'public_kid': 'aes192'
        },
        'aes256': {
            'provider': "local",
            'secret': os.urandom(32),
            'use': "enc",
            'public_kid': 'aes256'
        },
    }, keychain=keychain)
    assert keychain.get_public_key_versions('rsa')
    assert keychain.get_public_key_versions('p256')
    assert keychain.get_public_key_versions('p256k')


@pytest.fixture(scope='session')
def jwks(keychain):
    return jose.JSONWebKeySet.fromkeychain(keychain, [
        'rsa',
        'p256',
        'p256k',
        'p384',
        'p521',
        'ed448',
        'ed25519'
    ])


@pytest.fixture
def codec(keychain, jwks):
    assert keychain is not None
    return jose.PayloadCodec(
        signer=keychain,
        verifier=jwks,
        encrypter=keychain,
        decrypter=keychain
    )
