# pylint: skip-file
import time

import pytest

from ckms import jose
from ckms.utils import b64encode
from ckms.jose import ClaimSet



@pytest.mark.parametrize("claim", [
    "jti",
    "iss",
    "aud",
    "iat",
    "nbf",
    "exp"
])
def test_strict(claim):
    claims = ClaimSet.strict(
        iss="foo",
        aud="foo",
        ttl=600
    )
    assert getattr(claims, claim, None) is not None


def test_strict_sets_standardized_claim():
    claims = ClaimSet.strict(
        iss="foo",
        aud="foo",
        ttl=600,
        sub="foo"
    )
    assert claims.sub == "foo"


def test_strict_sets_private_claim():
    claims = ClaimSet.strict(
        iss="foo",
        aud="foo",
        ttl=600,
        foo="foo"
    )
    assert claims.extra['foo'] == "foo"


def test_dict_only_returns_present_claims():
    claims = ClaimSet(foo=1)
    data = claims.dict()
    assert len(data) == 1
    assert 'foo' in data


def test_json_only_returns_present_claims():
    claims = ClaimSet(foo=1)
    data = claims.json()
    assert data == '{"foo": 1}'


def test_get_attribute():
    claims = ClaimSet(foo=1)
    assert claims.foo == 1


def test_strict_sets_ttl():
    claims = ClaimSet.strict(
        iss="foo",
        aud="foo",
        ttl=600
    )
    assert claims.expires_in() == 600
    time.sleep(1)
    assert claims.expires_in() == 599


def test_serialize():
    claims = ClaimSet.frompayload(bytes(ClaimSet(iss='foo')))
    assert claims.iss == 'foo'


def test_frompayload():
    claims = ClaimSet.frompayload(b64encode(b'{"foo": 1}'))
    assert claims.extra.get('foo') == 1


def test_frompayload_raised_malformed():
    with pytest.raises(jose.MalformedPayload):
        ClaimSet.frompayload(b'f')


def test_invalid_exp_raises_invalidclaimtype():
    with pytest.raises(jose.InvalidClaimType):
        ClaimSet(exp='a')


def test_invalid_iat_raises_invalidclaimtype():
    with pytest.raises(jose.InvalidClaimType):
        ClaimSet(iat='a')


def test_invalid_nbf_raises_invalidclaimtype():
    with pytest.raises(jose.InvalidClaimType):
        ClaimSet(nbf='a')


def test_verify_raises_missing_required_claim():
    claims = ClaimSet(foo=1)
    with pytest.raises(jose.MissingRequiredClaim):
        claims.verify(required={'bar'})
    claims.verify(required={'foo'})


def test_validate_exp_raises():
    claims = ClaimSet(exp=1)
    with pytest.raises(jose.Expired):
        claims.verify(now=2)


def test_exp_from_expires_in():
    claims = ClaimSet()
    assert claims.expires_in(ttl=5, now=0) == 5
    assert claims.exp == 5


def test_validate_nbf_raises():
    claims = ClaimSet(nbf=2)
    with pytest.raises(jose.NotEffective):
        claims.verify(now=1)


def test_aud_is_set_when_not_provided():
    claims = ClaimSet()
    assert isinstance(claims.aud, set)


def test_aud_is_set_when_string():
    claims = ClaimSet(aud="foo")
    assert isinstance(claims.aud, set)
    assert claims.aud == {"foo"}


def test_aud_is_set_when_list():
    claims = ClaimSet(aud=["foo", "bar"])
    assert isinstance(claims.aud, set)
    assert claims.aud == {"foo", "bar"}


def test_aud_not_list_or_str_raises():
    with pytest.raises(jose.InvalidClaimType):
        claims = ClaimSet(aud=1)


def test_aud_must_be_all_string():
    with pytest.raises(jose.InvalidClaimType):
        claims = ClaimSet(aud=[1, "foo"])


def test_aud_must_intersect():
    claims = ClaimSet(aud=["foo"])
    with pytest.raises(jose.WrongAudience):
        claims.verify(audience={"bar"})
    claims.verify(audience={"foo", "bar"})


def test_verify_strict():
    with pytest.raises(jose.MissingRequiredClaim):
        claims = ClaimSet(iat=5, exp=10)
        claims.verify(strict=True, now=9)


def test_verify_max_age_exceeded():
    claims = ClaimSet(iat=1)
    with pytest.raises(jose.MaximumAgeExceeded):
        claims.verify(max_age=1, now=3)


def test_verify_max_age():
    claims = ClaimSet(iat=1)
    claims.verify(max_age=2, now=2)


def test_selfsigned():
    claims = ClaimSet(sub='foo', iss='foo')
    assert claims.is_selfsigned()