# pylint: skip-file
from multiprocessing.sharedctypes import Value
import os
import itertools
from ckms.jose.models.jsonwebkey import JSONWebKey

import jwcrypto.jwk
import pytest
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import ed448
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.kdf.concatkdf import ConcatKDFHash
from cryptography.hazmat.primitives import hashes

from ckms import jose

KEY_ALGORITHMS = [
    ("RSA", None, 2048),
    #("RSA", None, 3072),
    #("RSA", None, 4096),
    ("EC", "P-256", None),
    ("EC", "P-384", None),
    ("EC", "P-512", None),
    ("OKP", "Ed448", None),
    ("OKP", "Ed25519", None),
    ("oct", None, 32)
]


OKP_KEYS = [
    ("OKP", "Ed448", ed448.Ed448PrivateKey.generate()),
    ("OKP", "Ed25519", ed25519.Ed25519PrivateKey.generate()),
]

ASYMMETRIC_KEYS = [
    ("RSA", None, rsa.generate_private_key(public_exponent=65537, key_size=2048)),
    ("EC", "P-256", ec.generate_private_key(curve=ec.SECP256R1())),
    ("EC", "P-384", ec.generate_private_key(curve=ec.SECP384R1())),
    ("EC", "P-512", ec.generate_private_key(curve=ec.SECP521R1())),
    ("EC", "P-256K", ec.generate_private_key(curve=ec.SECP256K1())),
] + OKP_KEYS


SIGNING_ALGORITHMS = list(itertools.chain(
    itertools.product(
        ["RS256", "RS384", "RS512", "PS256", "PS384", "PS512"],
        [ASYMMETRIC_KEYS[0][2]]
    ),
    itertools.product(["ES256"], [ASYMMETRIC_KEYS[1][2]]),
    itertools.product(["ES384"], [ASYMMETRIC_KEYS[2][2]]),
    itertools.product(["ES512"], [ASYMMETRIC_KEYS[3][2]]),
    itertools.product(["ES256K"], [ASYMMETRIC_KEYS[4][2]]),
    itertools.product(["EdDSA"], [ASYMMETRIC_KEYS[5][2]]),
    itertools.product(["EdDSA"], [ASYMMETRIC_KEYS[6][2]]),
    itertools.product(["HS256", "HS384", "HS512"], [os.urandom(24)]),
))



ENCRYPTION_ALGORITHMS = itertools.chain(
    itertools.product(
        ["RSA1_5", "RSA-OAEP", "RSA-OAEP-256", "RSA-OAEP-384", "RSA-OAEP-512"],
        [ASYMMETRIC_KEYS[0][2]]
    ),
    itertools.product(
        ["A128KW", "A128GCM", "A128GCMKW", "A128CBC"],
        [os.urandom(16)]
    ),
    itertools.product(
        ["A192KW", "A192GCM", "A192GCMKW", "A192CBC"],
        [os.urandom(24)]
    ),
    itertools.product(
        ["A256KW", "A256GCM", "A256GCMKW", "A256CBC"],
        [os.urandom(32)]
    )
)



@pytest.mark.parametrize("kty,crv,private", ASYMMETRIC_KEYS)
def test_parse_asymmetric_public(kty, crv, private):
    public = private.public_key()
    jwk = jose.JSONWebKey.fromimpl(public)
    key = jwk.get_impl()
    assert jwk.kty == kty
    assert jwk.crv == crv
    if hasattr(public, 'public_numbers'):
        assert public.public_numbers() == key.public_numbers()
    else:
        r1 = public.public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw
        )
        r2 = public.public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw
        )
        assert r1 == r2


@pytest.mark.parametrize("kty,crv,private", ASYMMETRIC_KEYS)
def test_parse_asymmetric(kty, crv, private):
    jwk = jose.JSONWebKey.fromimpl(private)
    key = jwk.get_impl()
    assert jwk.kty == kty
    assert jwk.crv == crv
    if hasattr(private, 'private_numbers'):
        assert private.private_numbers() == key.private_numbers()
    else:
        r1 = private.private_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption()
        )
        r2 = private.private_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption()
        )
        assert r1 == r2


@pytest.mark.parametrize("kty,crv,private", ASYMMETRIC_KEYS)
def test_jwcrypto_jwk_compatibility(kty, crv, private):
    public = private.public_key()
    jwk = jose.JSONWebKey.fromimpl(public)
    jwk_jwcrypto = jwcrypto.jwk.JWK(**jwk.dict(exclude_defaults=True))
    if jwk.kty in {"EC", "OKP"}:
        assert jwk.crv == jwk_jwcrypto.key_curve
    assert jwk.kty == jwk_jwcrypto.kty


def test_fromkeychain_copies_properties(keychain):
    keys = keychain.get_public_key_versions(
        kid='rsa-enc',
        factory=lambda spec: jose.JSONWebKey.fromimpl(**spec.dict())
    )
    assert len(keys) == 1
    jwk = keys[0]
    assert jwk.kty == "RSA"
    assert jwk.use == "enc"





@pytest.mark.parametrize("alg,key", SIGNING_ALGORITHMS)
async def test_sign_and_verify(data, alg, key):
    jwk = JSONWebKey.fromimpl(key)
    sig = await jwk.sign(alg, data)
    assert await jwk.verify(alg, sig, data)


@pytest.mark.parametrize("alg,key", SIGNING_ALGORITHMS)
async def test_sign_and_verify_invalid(data, alg, key):
    jwk = JSONWebKey.fromimpl(key)
    sig = await jwk.sign(alg, data)
    assert not await jwk.verify(alg, sig, b'foo')


@pytest.mark.parametrize("alg,key", ENCRYPTION_ALGORITHMS)
async def test_encrypt_and_decrypt(data, alg, key):
    jwk = JSONWebKey.fromimpl(key)
    ct = await jwk.encrypt(alg, data)
    assert data == await jwk.decrypt(alg, ct)


@pytest.mark.parametrize("kty,crv,length", KEY_ALGORITHMS)
def test_generate_jwk(kty, crv, length):
    jwk = JSONWebKey.generate(kty, crv, length)


@pytest.mark.parametrize("kty", ["RSA", "oct"])
def test_generate_jwk_raises_typeerror(kty):
    with pytest.raises(TypeError):
        JSONWebKey.generate(kty, None, None)


def test_generate_jwk_raises_valueeeror():
    with pytest.raises(ValueError):
        JSONWebKey.generate("EC")


@pytest.mark.parametrize("kty,crv", [
    ("OKP", "X448"),
    ("OKP", "X25519")
])
async def test_jwk_key_exchange(kty, crv):
    kdf1 = ConcatKDFHash(
        algorithm=hashes.SHA256(),
        length=32,
        otherinfo=b'handshake data',
    )
    kdf2 = ConcatKDFHash(
        algorithm=hashes.SHA256(),
        length=32,
        otherinfo=b'handshake data',
    )
    k1 = JSONWebKey.generate(kty, crv)
    k2 = JSONWebKey.generate(kty, crv)
    s1, d1 = k1.exchange(k2.public, kdf1)
    s2, d2 = k2.exchange(k1.public, kdf2)

    assert s1.k == s2.k
    assert d1.k == d2.k