# pylint: skip-file
import typing

import httpx
import pytest
import jwt

import ckms
from ckms import jose
from ckms.jose.models import JSONWebKey
from ckms.jose.models import JSONWebKeySet
from .conftest import JOSE_SIGNING_ALGORITHMS


@pytest.mark.parametrize("jwks,expected", [
    (
        JSONWebKeySet(
            keys=[JSONWebKey(kty="RSA", alg="RSA-OAEP", use='enc')]
        ),
        ["RSA-OAEP"]
    ),
    (
        JSONWebKeySet(
            keys=[JSONWebKey(kty="RSA", use='enc')]
        ),
        ["RSA-OAEP", "RSA-OAEP-256", "RSA-OAEP-384", "RSA-OAEP-512", "RSA1_5"]
    ),
])
def test_get_encryption_algorithms(jwks: JSONWebKeySet, expected: list[str]):
    assert jwks.get_encryption_algorithms() == expected


@pytest.mark.parametrize("jwks,expected", [
    (
        JSONWebKeySet(
            keys=[JSONWebKey(kty="RSA", alg="RS256", use='sig')]
        ),
        ["RS256"]
    ),
    (
        JSONWebKeySet(
            keys=[
                JSONWebKey(kty="RSA", alg="RS256", use='sig'),
                JSONWebKey(kty="OKP", alg="EdDSA", crv="Ed448", use='sig'),
            ]
        ),
        ["EdDSA", "RS256"]
    ),
    (
        JSONWebKeySet(
            keys=[JSONWebKey(kty="RSA", use='sig')]
        ),
        ["PS256", "PS384", "PS512", "RS256", "RS384", "RS512"]
    ),
    (
        JSONWebKeySet(
            keys=[JSONWebKey(kty="EC", alg="ES256", use='sig')]
        ),
        ["ES256"]
    ),
    (
        JSONWebKeySet(
            keys=[JSONWebKey(kty="EC", crv="P-256", use='sig')]
        ),
        ["ES256"]
    ),
    (
        JSONWebKeySet(
            keys=[JSONWebKey(kty="OKP", alg="EdDSA", use='sig')]
        ),
        ["EdDSA"]
    ),
    (
        JSONWebKeySet(
            keys=[JSONWebKey(kty="OKP", use='sig')]
        ),
        ["EdDSA"]
    ),
    (
        JSONWebKeySet(
            keys=[JSONWebKey(kty="oct", alg="HS256", use='sig')]
        ),
        ["HS256"]
    ),
    (
        JSONWebKeySet(
            keys=[JSONWebKey(kty="oct", use='sig')]
        ),
        ["HS256", "HS384", "HS512"]
    ),
])
def test_get_signing_algorithms(
    jwks: JSONWebKeySet,
    expected: typing.List[str]
):
    assert jwks.get_signing_algorithms() == expected


def test_fromtags(keychain: ckms.Keychain):
    jwks = JSONWebKeySet.fromtags(keychain, ['oauth2'])
    assert len(jwks) == 3
    for name in {"rsa", "rsa-enc", "p256"}:
        meta = keychain.get_key_metadata(name)
        kid = meta.get_signing_kid()
        assert not jwks.has_key(name)
        assert jwks.has_key(kid)


def test_get_signing_keys(keychain: ckms.Keychain):
    jwks = JSONWebKeySet.fromtags(keychain, ['oauth2'])
    assert len(jwks) == 3
    assert len(jwks.get_signing_keys()) == 2


@pytest.mark.asyncio
async def test_verify_from_jws(codec, encoded_jwt, jwt, jwks):
    jwks = JSONWebKeySet(**jwks.dict())
    payload = await codec.decode_jws(encoded_jwt)
    assert isinstance(payload, jose.JSONWebSignature)
    assert await jwks.verify(payload)


@pytest.mark.asyncio
async def test_verify_from_jws_with_missing_key(codec, encoded_jwt, jwt):
    jwks = JSONWebKeySet()
    payload = await codec.decode_jws(encoded_jwt)
    assert isinstance(payload, jose.JSONWebSignature)
    assert not await jwks.verify(payload)


@pytest.mark.parametrize("kid,alg", JOSE_SIGNING_ALGORITHMS)
@pytest.mark.asyncio
async def test_verify_from_jws_without_kid(codec, keychain, jwks, kid, alg):
    if kid == 'hmac':
        pytest.skip()
    jwks = JSONWebKeySet(**jwks.dict())
    public, private = keychain.get_key_material(kid)
    token = jwt.encode({'foo': 'bar'}, private, algorithm=alg)
    payload = await codec.decode_jws(token)
    assert isinstance(payload, jose.JSONWebSignature)
    assert not payload._signatures[0].kid
    assert await jwks.verify(payload)


@pytest.mark.asyncio
async def test_spoof_header_hmac(keychain, codec):
    public, private = keychain.get_key_material('rsa')
    jwks = JSONWebKeySet(
        keys=[
            JSONWebKey.frompem(public)
        ]
    )

    from authlib.jose import jwt
    token = jwt.encode({'alg': "HS256"}, {}, public)
    payload = await codec.decode_jws(token)
    assert not await jwks.verify(payload)


def test_jwks_factory(keychain):
    jwks = jose.JSONWebKeySet.fromkeychain(keychain, ['rsa', 'p256', 'p256k'])
    assert len(jwks) == 3


def test_jwks_to_dict(keychain):
    jwks = jose.JSONWebKeySet.fromtags(keychain, ['oauth2'])
    assert len(jwks.dict()['keys']) == 3


def test_jwks_fromtags(keychain):
    jwks = jose.JSONWebKeySet.fromtags(keychain, ['oauth2'])
    assert len(jwks) == 3
    assert jwks.has_public_key(keychain.get_public_kid('rsa'))
    assert jwks.has_public_key(keychain.get_public_kid('rsa-enc'))
    assert jwks.has_public_key(keychain.get_public_kid('p256'))


def test_jwks_fromtags_preserves_order(keychain):
    jwks = jose.JSONWebKeySet.fromtags(keychain, ['oauth2'])
    assert len(jwks) == 3
    assert jwks[0].use == "enc"
    assert jwks[1].use == "sig"


def test_jwks_factory_sets_public_kids(keychain):
    jwks = jose.JSONWebKeySet.fromkeychain(keychain, ['rsa', 'p256', 'p256k'])
    keys = jwks.get_key_identifiers()
    for name in ['rsa', 'p256', 'p256k']:
        meta = keychain.get_key_metadata(name)
        assert meta.get_signing_kid() in keys


@pytest.mark.asyncio
async def test_jwks_clone(keychain):
    jwks1 = jose.JSONWebKeySet.fromkeychain(keychain, ['rsa', 'p256', 'p256k'])
    jwks2 = await jose.JSONWebKeySet.fromjwks(JSONWebKeySet(**jwks1.dict()))
    assert len(jwks1) == len(jwks2)


@pytest.mark.asyncio
async def test_jwks_verify(keychain, data):
    private = keychain.get('rsa')
    jwks = jose.JSONWebKeySet.fromkeychain(keychain, ['rsa'])
    sig = await keychain.sign(data, using='rsa', algorithm='RS256')
    assert await jwks.verify(sig, data, using=private.public_keys[0].kid, algorithm='RS256')


@pytest.mark.asyncio
async def test_jwks_verify_failure(keychain, data):
    private = keychain.get('rsa')
    jwks = jose.JSONWebKeySet.fromkeychain(keychain, ['rsa'])
    sig = await keychain.sign(data, using='rsa', algorithm='RS256')
    assert not await jwks.verify(sig, data + b'foo', using=private.public_keys[0].kid, algorithm='RS256')


@pytest.mark.asyncio
async def test_jwks_verify_without_kid(keychain, data):
    private = keychain.get('rsa')
    jwks = jose.JSONWebKeySet.fromkeychain(keychain, ['rsa'])
    sig = await keychain.sign(data, using='rsa', algorithm='RS256')
    assert await jwks.verify(sig, data, algorithm='RS256')


@pytest.mark.asyncio
async def test_jwks_verify_without_kid_failure(keychain, data):
    private = keychain.get('rsa')
    jwks = jose.JSONWebKeySet.fromkeychain(keychain, ['rsa'])
    sig = await keychain.sign(data, using='rsa', algorithm='RS256')
    assert not await jwks.verify(sig, data + b'foo', algorithm='RS256')


@pytest.mark.asyncio
async def test_create_http_client():
    jwks = jose.JSONWebKeySet()
    assert isinstance(jwks.http_client_factory("foo"), httpx.AsyncClient)
