# pylint: skip-file
import json

import pyexpat
import pytest

from ckms import jose
from ckms.jose import exceptions
from ckms.jose import JSONWebToken
from ckms.utils import b64decode
from .conftest import SIGNING_KEYS


def test_single_aud_is_serialized_to_string():
    jwt = jose.JSONWebToken(aud=["foo"])
    claims = json.loads(b64decode(bytes(jwt)))
    assert claims.get('aud') == "foo"


def test_jwt_to_dict():
    jwt = jose.JSONWebToken(foo=1)
    assert dict(jwt) == {'foo': 1}


def test_jwt_serialize_and_deserialize():
    t1 = jose.JSONWebToken(foo=1)
    buf = bytes(t1)
    t2 = jose.JSONWebToken.fromb64(buf)
    assert t1.get('foo') == t2.get('foo')


@pytest.mark.parametrize("claim,value", [
    ('exp', 'foo'),
    ('iat', 'foo'),
    ('nbf', 'foo'),
])
def test_invalid_values_raises_invalidclaimtype(claim, value):
    jwt = JSONWebToken(**{claim: value})
    with pytest.raises(exceptions.InvalidClaimType):
        getattr(jwt, claim)


@pytest.mark.asyncio
async def test_validate_exp_raises():
    jwt = JSONWebToken(exp=1)
    with pytest.raises(exceptions.Expired):
        await jwt.validate(now=2)


@pytest.mark.asyncio
async def test_exp_from_expires_in():
    jwt = JSONWebToken()
    jwt.expires_in(ttl=5, now=0)
    assert jwt.exp == 5


@pytest.mark.asyncio
async def test_validate_nbf_raises():
    jwt = JSONWebToken(nbf=2)
    with pytest.raises(exceptions.NotEffective):
        await jwt.validate(now=1)

@pytest.mark.asyncio
async def test_validate_iat_raises():
    jwt = JSONWebToken(iat=2)
    with pytest.raises(exceptions.InvalidClaim):
        await jwt.validate(now=1)


def test_aud_is_set_when_not_provided():
    jwt = JSONWebToken()
    assert isinstance(jwt.aud, set)


def test_aud_is_set_when_string():
    jwt = JSONWebToken(aud="foo")
    assert isinstance(jwt.aud, set)
    assert jwt.aud == {"foo"}


def test_aud_is_set_when_list():
    jwt = JSONWebToken(aud=["foo", "bar"])
    assert isinstance(jwt.aud, set)
    assert jwt.aud == {"foo", "bar"}


def test_aud_not_list_or_str_raises():
    jwt = JSONWebToken(aud=1)
    with pytest.raises(exceptions.InvalidClaimType):
        jwt.aud


def test_aud_must_be_all_string():
    jwt = JSONWebToken(aud=[1, "foo"])
    with pytest.raises(exceptions.InvalidClaim):
        jwt.validate_claims()


def test_aud_must_intersect():
    jwt = JSONWebToken(aud=["foo"])
    with pytest.raises(exceptions.WrongAudience):
        jwt.validate_claims(audience={"bar"})
    jwt.validate_claims(audience={"foo", "bar"})