# pylint: skip-file
# type: ignore
from ckms.algorithms import algorithm
from ckms.jose.jsonwebtoken import JSONWebToken
import pytest

from ckms import jose
from tests.jose.conftest import payload


#@pytest.mark.asyncio
#async def test_reject_token_type(codec: jose.PayloadCodec):
#    payload = codec.encode({'foo': 'bar'})
#    token = bytes(await payload.sign('RS256', 'rsa', header={'typ': 'foo'}))
#    with pytest.raises(jose.exceptions.UnsupportedTokenType):
#        await codec.decode(
#            token=token,
#            accept_token={"jwt"}
#        )
#
#
#@pytest.mark.asyncio
#async def test_accept_token_type(codec: jose.PayloadCodec):
#    payload = codec.encode({'foo': 'bar'})
#    token = bytes(await payload.sign('RS256', 'rsa'))
#    jwt = await codec.decode(
#        token=token,
#        accept_token={"jwt"}
#    )
#    assert isinstance(jwt, JSONWebToken)


@pytest.mark.asyncio
async def test_encrypto_fixed_algorithm(codec: jose.PayloadCodec):
    alg = codec.get_key_algorithm('rsa-fixed-enc')
    assert alg is not None
    payload = codec.encode({'foo': 'bar'})
    token = await payload.encrypt(using='rsa-fixed-enc')
    header, claims = codec.introspect(token)
    assert header.alg == alg


@pytest.mark.asyncio
async def test_sign_fixed_algorithm(codec: jose.PayloadCodec):
    alg = codec.get_key_algorithm('rsa-fixed')
    payload = codec.encode({'foo': 'bar'})
    token = await payload.sign(using='rsa-fixed')
    header, claims = codec.introspect(token)
    assert header.alg == alg


@pytest.mark.asyncio
async def test_introspect_binary(codec: jose.PayloadCodec):
    payload = codec.encode(b'Hello world!')
    token = bytes(await payload.sign(algorithm='RS256', using='rsa'))
    _, claims = jose.PayloadCodec.introspect(token)
    assert claims is None


@pytest.mark.asyncio
async def test_introspect_jws_with_model(codec: jose.PayloadCodec):
    class CustomClaimSet(jose.ClaimSet):
        foo: str

    claims: CustomClaimSet
    payload = codec.encode(jose.ClaimSet.strict(iss='foo', aud='foo', foo='foo', ttl=30))
    token = bytes(await payload.sign(algorithm='RS256', using='rsa'))
    _, claims = jose.PayloadCodec.introspect(token, model=CustomClaimSet)
    assert isinstance(claims, CustomClaimSet), type(claims)
    assert claims is not None
    assert claims.foo == 'foo'


@pytest.mark.asyncio
async def test_introspect_jws(codec: jose.PayloadCodec):
    payload = codec.encode(jose.ClaimSet.strict(iss='foo', aud='foo', ttl=30))
    token = bytes(await payload.sign(algorithm='RS256', using='rsa'))
    _, claims = jose.PayloadCodec.introspect(token)
    assert claims is not None
    assert claims.iss == 'foo'


@pytest.mark.asyncio
async def test_codec_encode_claimset(codec: jose.PayloadCodec):
    payload = codec.encode(jose.ClaimSet.strict(iss='foo', aud='foo', ttl=30))
    jwt = await codec.decode(bytes(await payload.sign(algorithm='RS256', using='rsa')), validate=False)
    assert jwt.claims['aud'] == 'foo'
    assert jwt.claims['iss'] == 'foo'


@pytest.mark.asyncio
async def test_codec_encode_dict(codec: jose.PayloadCodec):
    payload = codec.encode({'foo': 1})
    jwt = await codec.decode(bytes(await payload.sign(algorithm='RS256', using='rsa')))
    assert jwt.claims['foo'] == 1


@pytest.mark.asyncio
async def test_codec_encode_jws_with_overriden_typ(codec: jose.PayloadCodec):
    signed = await codec.encode(JSONWebToken())\
        .sign(algorithm='RS256', using='rsa', typ="at+jwt")
    jws = await codec.decode_jws(bytes(signed))
    assert jws.typ == "at+jwt"


@pytest.mark.asyncio
async def test_codec_encode_jws_with_overriden_can_deserialize_jwt(codec: jose.PayloadCodec):
    signed = await codec.encode(JSONWebToken())\
        .sign(algorithm='RS256', using='rsa', typ="at+jwt")
    jwt = await codec.decode(bytes(signed))
    assert isinstance(jwt, jose.JSONWebToken)


@pytest.mark.asyncio
async def test_codec_encode_and_sign_jwt(codec):
    jwt = jose.JSONWebToken(foo=1)
    payload = codec.encode(jwt)
    jws = await payload.sign(algorithm='RS256', using='rsa')
    decoded = await codec.decode(bytes(jws))
    assert isinstance(decoded, jose.JSONWebToken), repr(decoded)


@pytest.mark.asyncio
async def test_codec_decode_jws(codec, encoded_jwt, jwt):
    payload = await codec.decode_jws(encoded_jwt)
    assert isinstance(payload, jose.JSONWebSignature)
    assert payload.get_claimset() is not None


@pytest.mark.asyncio
async def test_codec_decode_jws_jwt(codec, encoded_jwt, jwt):
    payload = await codec.decode(encoded_jwt)
    assert jwt.claims.get('foo') == payload.claims.get('foo')


@pytest.mark.asyncio
async def test_codec_decode_jws_payload(codec, encoded_payload, payload):
    decoded = await codec.decode(encoded_payload)
    assert decoded == payload


@pytest.mark.asyncio
async def test_codec_decode_jwe_jwt(codec, encrypted_jwt, jwt):
    decrypted = await codec.decode(encrypted_jwt)
    assert isinstance(jwt, jose.JSONWebToken)
    assert jwt.claims.get('foo') == decrypted.claims.get("foo")


@pytest.mark.asyncio
async def test_codec_decode_jwe_payload(codec, encrypted_payload, payload):
    decrypted = await codec.decode(encrypted_payload)
    assert decrypted == payload


@pytest.mark.asyncio
async def test_deserialize_unsecured(keychain, codec):
    jwt = jose.JSONWebToken(foo=1)
    t1 = codec.encode(jwt)
    jws = await t1.sign(algorithm="none", using=None)
    token = bytes(jws)
    t2 = await codec.decode(token, allow_none=True)
    assert jwt.claims.get('foo') == t2.claims.get('foo')


@pytest.mark.asyncio
async def test_deserialize_unsecured_raises_when_not_allowed(keychain, codec):
    jwt = jose.JSONWebToken(foo=1)
    t1 = codec.encode(jwt)
    jws = await t1.sign(algorithm="none", using=None)
    token = bytes(jws)
    with pytest.raises(jose.AlgorithmNotAllowed):
        await codec.decode(token, allow_none=False)


@pytest.mark.asyncio
async def test_deserialize_nested_jws(keychain, codec):
    payload = codec.encode(jose.JSONWebToken(foo=1))
    jws = await payload.sign(algorithm='RS256', using='rsa')
    jwe = await jws.encrypt(algorithm='RSA-OAEP-256', using='rsa')
    jwt = await codec.decode(bytes(jwe))


@pytest.mark.asyncio
async def test_deserialize_jws_payload(keychain, codec):
    payload = codec.encode(b"Hello world!")
    jws = await payload.sign(algorithm='RS256', using='rsa')
    assert await codec.decode(bytes(jws)) == b"Hello world!"


@pytest.mark.asyncio
async def test_deserialize_jwe_payload(keychain, codec):
    payload = codec.encode(b"Hello world!")
    jwe = await payload.encrypt(algorithm='RSA-OAEP', using='rsa')
    assert await codec.decode(bytes(jwe)) == b"Hello world!"