# pylint: skip-file
import pytest
import pytest_asyncio

import ckms
import ckms.jose
from ckms.jose.models import JSONWebKeySet


@pytest_asyncio.fixture # type: ignore
async def local_keychain():
    keychain = ckms.Keychain()
    keychain.register({
        'sig': {
            'provider': "local",
            'kty': "OKP",
            'crv': "Ed25519"
        }
    })
    await keychain.setup()
    return keychain


@pytest_asyncio.fixture # type: ignore
async def external_codec(local_keychain: ckms.Keychain):
    keychain = ckms.Keychain()
    keychain.register({
        'sig': {
            'provider': "local",
            'kty': "OKP",
            'crv': "Ed25519"
        }
    })
    await keychain.setup()
    return ckms.jose.PayloadCodec(
        signer=keychain,
        encrypter=JSONWebKeySet.fromkeychain(local_keychain)
    )


@pytest_asyncio.fixture # type: ignore
async def untrusted_codec():
    keychain = ckms.Keychain()
    keychain.register({
        'sig': {
            'provider': "local",
            'kty': "OKP",
            'crv': "Ed25519"
        }
    })
    await keychain.setup()
    return ckms.jose.PayloadCodec(signer=keychain)


@pytest_asyncio.fixture # type: ignore
async def external_jwks(external_codec: ckms.jose.PayloadCodec):
    return JSONWebKeySet.fromkeychain(external_codec.signer)


@pytest.fixture
def issuer():
    return "https://ckms.localhost"


@pytest_asyncio.fixture # type: ignore
async def issuers(external_jwks: JSONWebKeySet, issuer: str):
    issuers = ckms.jose.TrustedIssuers()
    issuers.trust(issuer, external_jwks)
    await issuers.setup()
    return issuers


@pytest.mark.asyncio
async def test_verify_signed_token_from_issuer(
    external_codec: ckms.jose.PayloadCodec,
    issuers: ckms.jose.TrustedIssuers,
    issuer: str
):
    payload = external_codec.encode(
        payload={'iss': issuer}
    )
    jws = await payload.sign(
        algorithm='EdDSA',
        using='sig'
    )
    assert await issuers.verify_token(bytes(jws))


@pytest.mark.asyncio
async def test_verify_signed_token_from_untrusted_issuer_with_trusted_key(
    external_codec: ckms.jose.PayloadCodec,
    issuers: ckms.jose.TrustedIssuers,
    issuer: str
):
    payload = external_codec.encode(
        payload={'iss': "https://trudy.localhost"}
    )
    jws = await payload.sign(
        algorithm='EdDSA',
        using='sig'
    )
    with pytest.raises(issuers.UntrustedIssuer):
        await issuers.verify_token(bytes(jws))


@pytest.mark.asyncio
async def test_verify_signed_token_from_trusted_issuer_with_untrusted_key(
    untrusted_codec: ckms.jose.PayloadCodec,
    issuers: ckms.jose.TrustedIssuers,
    issuer: str
):
    payload = untrusted_codec.encode(
        payload={'iss': issuer}
    )
    jws = await payload.sign(
        algorithm='EdDSA',
        using='sig'
    )
    with pytest.raises(issuers.InvalidSignature):
        await issuers.verify_token(bytes(jws))


@pytest.mark.asyncio
async def test_verify_signed_token_from_issuer_with_missing_iss_claim(
    external_codec: ckms.jose.PayloadCodec,
    issuers: ckms.jose.TrustedIssuers,
    issuer: str
):
    payload = external_codec.encode(
        payload={}
    )
    jws = await payload.sign(
        algorithm='EdDSA',
        using='sig'
    )
    with pytest.raises(issuers.UntrustedIssuer):
        assert await issuers.verify_token(bytes(jws))