"""Declares :class:`Payload`"""
import typing
import warnings

from ckms.utils import b64encode
from .jsonwebencryption import JSONWebEncryption
from .jsonwebsignature import JSONWebSignature
from .jsonwebtoken import JSONWebToken


ENCODED_TYPES = (
    JSONWebEncryption,
    JSONWebSignature,
    JSONWebToken,
)


class Payload:
    __module__: str = 'ckms.jose'

    def __init__(self,
        codec: 'PayloadCodec',
        payload: typing.Union[bytes, JSONWebSignature, JSONWebToken],
        claims: typing.Optional[dict[str, typing.Any]] = None
    ):
        self._codec = codec
        self._payload = payload
        self._claims = claims or {}

    async def encrypt(
        self,
        *,
        algorithm: str | None = None,
        using: str | None = None,
        encryption: typing.Optional[str] = None
    ) -> 'Payload':
        """Encrypt the :class:`Payload` instance."""
        if using is None:
            raise ValueError("The `using` parameter is required.")
        algorithm = algorithm or self._codec.get_key_algorithm(using)
        if algorithm is None:
            raise ValueError(
                "The `algorithm` parameter is None and the key `using` "
                "does not specify a default algorithm."
            )

        return Payload(
            codec=self._codec,
            payload=await self._codec.encrypt(
                algorithm=algorithm,
                using=using,
                payload=self,
                encryption=encryption,
                claims=self.get_header_claims()
            )
        )

    def get_header_claims(self) -> dict:
        """Return a dictionary containing additional claims to add to
        the JOSE header.
        """
        claims = self._claims
        if isinstance(self._payload, JSONWebToken):
            claims['typ'] = "JWT"
        elif isinstance(self._payload, (JSONWebEncryption, JSONWebSignature)):
            # If the payload is either JSONWebEncryption or JSONWebSignature,
            # then we are creating a nested JWT and thus should set the `cty`
            # claim.
            claims['cty'] = "JWT"
        return claims

    async def sign(
        self,
        *, 
        algorithm: str | None = None,
        using: str | None = None,
        typ: str | None = None,
        claims: dict[str, typing.Any] | None = None,
        header: dict[str, typing.Any] | None = None
    ) -> 'Payload':
        """Sign the :class:`Payload` instance."""
        if using is None and algorithm != 'none':
            raise ValueError("The `using` parameter is required.")
        algorithm = algorithm or self._codec.get_key_algorithm(using)
        if algorithm is None:
            raise ValueError(
                "The `algorithm` parameter is None and the key `using` "
                "does not specify a default algorithm."
            )
        claims = claims or {}
        header = header or {}
        if typ is not None:
            warnings.warn("The `typ` parameter is deprecated", DeprecationWarning)
            header.setdefault('typ', typ)
        if claims:
            warnings.warn("The `claims` parameter is deprecated", DeprecationWarning)
        if isinstance(self._payload, JSONWebToken):
            header.setdefault('typ', "jwt")
        jws = JSONWebSignature(payload=self)
        await self._codec.sign_jws(algorithm, using, jws, {**header, **claims})
        assert jws.is_signed() # nosect
        return Payload(
            codec=self._codec,
            payload=jws
        )

    def __bytes__(self) -> bytes:
        # For JWS/JWE, the payload consists of segments that are
        # already encoded. For raw data, the encoding needs to be
        # applied.
        if isinstance(self._payload, ENCODED_TYPES):
            payload = bytes(self._payload)
        else:
            payload = b64encode(bytes(self._payload))
        return payload

    def __str__(self) -> str:
        return bytes.decode(bytes(self), 'ascii')
